#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Standard IDAES pressure changer model.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Changing existing config block attributes
# pylint: disable=protected-access

# Import Python libraries


# Import Pyomo libraries
from pyomo.environ import (
    Block,
    value,
    Var,
    Expression,
    Constraint,
    Reference,
    check_optimal_termination,
    Reals,
    Param,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    EnergyBalanceType,
    MomentumBalanceType,
    MaterialBalanceType,
    ProcessBlockData,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.exceptions import PropertyNotSupportedError, InitializationError
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.initialization import SingleControlVolumeUnitInitializer
from idaes.core.util import to_json, from_json, StoreSpec
from idaes.core.util.math import smooth_max, safe_sqrt, sqrt, smooth_min
from pyomo.environ import units as pyunits


__author__ = "Emmanuel Ogbe, Andrew Lee"
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("TurbineBase")
class TurbineBaseData(UnitModelBlockData):
    """
    Standard Compressor/Expander Unit Model Class
    """

    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be
constructed, **default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Phase equilibrium construction flag",
            doc="""Indicates whether terms for phase equilibrium should be
constructed, **default** = False.
**Valid values:** {
**True** - include phase equilibrium terms
**False** - exclude phase equilibrium terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property
calculations, **default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "calculation_method",
        ConfigValue(
            default='isentropic',
            domain=In(["isentropic", "simple_willans", "part_load_willans", "Tsat_willans", "BPST_willans", "CT_willans"]),
            description="Calculation method used to model mechanical work",
            doc="""Property parameter object used to define property
calculations, **default** - 'isentropic'.
**Valid values:** {
**isentropic** - default method, uses isentropic efficiency to determine work
**simple_willans** - simple willans line requring slope, intercept and max molar flow.
**part_load_willans** - willans line with part load correction, a, b, c and max molar flow as parameters
**Tsat_willans** - willans line with part load correction using saturation temperature. Requires only max molar flow
**BPST_willans** - back pressure willans line with part load correction using pressure difference, requires max molar flow.
**CT_willans** - condensing turbine willans line with part load correction using pressure difference, requires max molar flow.
}""",
        ),
    )

    def build(self):
        """

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build
        super().build()

        # Add a control volume to the unit including setting up dynamics.
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        # Add geometry variables to control volume
        if self.config.has_holdup:
            self.control_volume.add_geometry()

        # Add inlet and outlet state blocks to control volume
        self.control_volume.add_state_blocks(
            has_phase_equilibrium=self.config.has_phase_equilibrium
        )

        # Add mass balance
        # Set has_equilibrium is False for now
        # TO DO; set has_equilibrium to True
        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=self.config.has_phase_equilibrium,
        )

        # Add energy balance
        eb = self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_work_transfer=True
        )

        # add momentum balance
        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Set Unit Geometry and holdup Volume
        if self.config.has_holdup is True:
            self.volume = Reference(self.control_volume.volume[:])

        self.work_mechanical = Reference(self.control_volume.work[:])
        # self.work_mechanical.fix(-1000e3)
        # self.work_mechanical.unfix()


        # Add Momentum balance variable 'deltaP'
        self.deltaP = Reference(self.control_volume.deltaP[:])

        # Performance Variables
        self.ratioP = Var(self.flowsheet().time, initialize=1.0, doc="Pressure Ratio")

        # Pressure Ratio
        @self.Constraint(self.flowsheet().time, doc="Pressure ratio constraint")
        def ratioP_calculation(self, t):
            return (
                self.ratioP[t] * self.control_volume.properties_in[t].pressure
                == self.control_volume.properties_out[t].pressure
            )

        units_meta = self.control_volume.config.property_package.get_metadata()

        # Get indexing sets from control volume
        # Add isentropic variables
        self.efficiency_isentropic = Var(
            self.flowsheet().time,
            initialize=0.5,
            doc="Efficiency with respect to an isentropic process [-]",
        )

        # self.delta_h_is = Var(
        #     self.flowsheet().time,
        #     initialize=-100e3,
        #     doc="Enthalpy difference input to unit if isentropic process",
        #     units=units_meta.get_derived_units("energy") / units_meta.get_derived_units("amount"),
        # )
        # self.delta_h_act = Var(
        #     self.flowsheet().time,
        #     initialize=-100e3,
        #     doc="Enthalpy difference input to unit for actual process",
        #     units=units_meta.get_derived_units("energy") / units_meta.get_derived_units("amount"),
        # )
        # self.work_isentropic = Var(
        #     self.flowsheet().time,
        #     initialize=-100e3,
        #     doc="Work input to unit if isentropic process",
        #     units=units_meta.get_derived_units("power"),
        # )

        # Add motor/electrical work and efficiency variable
        self.efficiency_motor = Var(
            self.flowsheet().time,
            initialize=1.0,
            doc="Motor efficiency converting shaft work to electrical work [-]",
            )
        
        self.work_electrical = Var(
            self.flowsheet().time,
            initialize=1.0,
            doc="Electrical work of a turbine [-]",
            units=units_meta.get_derived_units("power")
            )

        # Add willans line parameters
        if 'willans' in self.config.calculation_method:
            self.willans_slope = Var(
                self.flowsheet().time,
                initialize=100,
                doc="Slope of willans line",
                units=units_meta.get_derived_units("energy") / units_meta.get_derived_units("amount"),
            )

            self.willans_intercept = Var(
                self.flowsheet().time,
                initialize=-100,
                doc="Intercept of willans line",
                units=units_meta.get_derived_units("power"),
            )

            self.willans_max_mol = Var(
                self.flowsheet().time,
                initialize=1.0,
                doc="Max molar flow of willans line",
                units=units_meta.get_derived_units("amount") / units_meta.get_derived_units("time"),
            )

            if self.config.calculation_method in ["part_load_willans", "Tsat_willans", "BPST_willans", "CT_willans"]:
                self.willans_a = Var(
                    self.flowsheet().time,
                    initialize=1.0,
                    doc="Willans a coefficient",
                )

                self.willans_b = Var(
                    self.flowsheet().time,
                    initialize=1.0,
                    doc="Willans b coefficient",
                    units=units_meta.get_derived_units("power")
                )

                self.willans_efficiency = Var(
                    self.flowsheet().time,
                    initialize=1.0,
                    doc="Willans efficiency",
                )

        # Build isentropic state block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = self.config.has_phase_equilibrium
        tmp_dict["defined_state"] = False

        self.properties_isentropic = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="isentropic properties at outlet", **tmp_dict
        )

        # Connect isentropic state block properties
        @self.Constraint(
            self.flowsheet().time, doc="Pressure for isentropic calculations"
        )
        def isentropic_pressure(self, t):
            return (
                self.properties_isentropic[t].pressure
                == self.control_volume.properties_out[t].pressure
            )

        # This assumes isentropic composition is the same as outlet
        self.add_state_material_balances(
            self.config.material_balance_type,
            self.properties_isentropic,
            self.control_volume.properties_out,
        )

        # This assumes isentropic entropy is the same as inlet
        @self.Constraint(self.flowsheet().time, doc="Isentropic assumption")
        def isentropic(self, t):
            return (
                self.properties_isentropic[t].entr_mol
                == self.control_volume.properties_in[t].entr_mol
            )
    
        self.add_isentropic_work_definition()
        
        if 'willans' in self.config.calculation_method: 
            # Write isentropic efficiency eqn
            self.add_willans_line_relationship() 

            if self.config.calculation_method in ["part_load_willans", "Tsat_willans", "BPST_willans", "CT_willans"]:
                if self.config.calculation_method == "Tsat_willans": # use published values and dTsat to calculate willans a,b,c
                    self.calculate_Tsat_willans_parameters()

                elif self.config.calculation_method == "BPST_willans":
                    self.calculate_BPST_willans_parameters()
                
                elif self.config.calculation_method == "CT_willans":
                    self.calculate_CT_willans_parameters()

                # calculate slope and intercept
                self.calculate_willans_coefficients()

        self.add_mechanical_and_isentropic_work_definition()
        self.add_electrical_work_definition()
       
    def calculate_CT_willans_parameters(self):
        # a parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans CT a calculation"
        )
        def willans_CT_a_calculation(self, t):
            return self.willans_a[t] == 1.288464 - 0.0015185 * (self.control_volume.properties_in[t].pressure / 1e5) / pyunits.Pa - 0.33415834 * (self.control_volume.properties_out[t].pressure / 1e5) / pyunits.Pa
        
        # b parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans CT b calculation"
        )
        def willans_CT_b_calculation(self, t):
            return self.willans_b[t] == (-437.7746025 + 29.00736723 * (self.control_volume.properties_in[t].pressure / 1e5) / pyunits.Pa + 10.35902331 * (self.control_volume.properties_out[t].pressure / 1e5) / pyunits.Pa) * 1000 * pyunits.W
        
        # c parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans CT efficiency calculation"
        )
        def willans_CT_efficiency_calculation(self, t):
            return self.willans_efficiency[t] == 1 / ((0.07886297 + 0.000528327 * (self.control_volume.properties_in[t].pressure / 1e5) / pyunits.Pa - 0.703153891 * (self.control_volume.properties_out[t].pressure / 1e5) / pyunits.Pa) + 1)      

    def calculate_BPST_willans_parameters(self):
        # a parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans BPST a calculation"
        )
        def willans_BPST_a_calculation(self, t):
            return self.willans_a[t] == 1.18795366 - 0.00029564 * (self.control_volume.properties_in[t].pressure / 1e5) / pyunits.Pa + 0.004647288 * (self.control_volume.properties_out[t].pressure / 1e5) / pyunits.Pa
        

        # b parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans BPST b calculation"
        )
        def willans_BPST_b_calculation(self, t):
            return self.willans_b[t] == (449.9767142 + 5.670176939 * (self.control_volume.properties_in[t].pressure / 1e5) / pyunits.Pa - 11.5045814 * (self.control_volume.properties_out[t].pressure / 1e5) / pyunits.Pa) * 1000 * pyunits.W
        

        # c parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans BPST c calculation"
        )
        def willans_BPST_efficiency_calculation(self, t):
            return self.willans_efficiency[t] == 1 / ((0.205149333 - 0.000695171 * (self.control_volume.properties_in[t].pressure / 1e5) / pyunits.Pa + 0.002844611 * (self.control_volume.properties_out[t].pressure / 1e5) / pyunits.Pa) + 1)       

    def calculate_Tsat_willans_parameters(self):
        # a parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans Tsat a calculation"
        )
        def willans_Tsat_a_calculation(self, t):
            return self.willans_a[t] == (1.155 + 0.000538 * (self.control_volume.properties_in[t].temperature_sat - self.control_volume.properties_out[t].temperature_sat) / pyunits.K)

        # b parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans Tsat b calculation"
        )
        def willans_Tsat_b_calculation(self, t):
            return self.willans_b[t] == (0 + 4.23 * (self.control_volume.properties_in[t].temperature_sat - self.control_volume.properties_out[t].temperature_sat) / pyunits.K)*1000 * pyunits.W
        
        # c parameter
        @self.Constraint(
                self.flowsheet().time, doc="Willans Tsat efficiency calculation"
        )
        def willans_Tsat_c_calculation(self, t):
            return self.willans_efficiency[t] == 0.83333

    def calculate_willans_coefficients(self):
        # Calculate willans coefficients
            @self.Constraint(
                self.flowsheet().time, doc="Willans slope calculation"
            )
            def willans_slope_calculation(self, t):
                return self.willans_slope[t] == 1 / (self.willans_efficiency[t] * self.willans_a[t]) * ((self.control_volume.properties_in[t].enth_mol - self.properties_isentropic[t].enth_mol) - self.willans_b[t] / self.willans_max_mol[t])
            
            
            @self.Constraint(
                self.flowsheet().time, doc="Willans intercept calculation"
            )
            def willans_intercept_calculation(self, t):
                return self.willans_intercept[t] == ((1 - self.willans_efficiency[t]) / (self.willans_efficiency[t] * self.willans_a[t])) * ((self.control_volume.properties_in[t].enth_mol - self.properties_isentropic[t].enth_mol) * self.willans_max_mol[t] - self.willans_b[t])
    
    def add_mechanical_and_isentropic_work_definition(self):
        
        if 'willans' in self.config.calculation_method: 
            self.efficiency_isentropic = Expression(
                self.flowsheet().time,
                rule=lambda b, t: (
                    b.delta_h_act[t] / b.delta_h_is[t]
                )
            )
        else:
            # Mechanical work
            @self.Constraint(
                self.flowsheet().time, doc="Isentropic and mechanical work relationship"
            )
            def isentropic_and_mechanical_work_eq(self, t):
                    return self.work_mechanical[t] == (
                        self.delta_h_is[t] * self.control_volume.properties_in[t].flow_mol * self.efficiency_isentropic[t]
                    )
            
    def add_willans_line_relationship(self):
        @self.Constraint(
            self.flowsheet().time, doc="Willans line and mechanical work relationship"
        )
        def willans_line_eq(self, t):            
            eps = 0.0001  # smoothing parameter; smaller = closer to exact max, larger = smoother
            return self.work_mechanical[t] == smooth_min(
                -(self.willans_slope[t] * self.control_volume.properties_in[t].flow_mol - self.willans_intercept[t]) / (self.willans_slope[t] * self.willans_max_mol[t]),
                0.0,
                eps
                ) * (self.willans_slope[t] * self.willans_max_mol[t])
         
    def add_electrical_work_definition(self):
        # Electrical work
        @self.Constraint(
            self.flowsheet().time, doc="Calculate electrical work of turbine"
        )
        def electrical_energy_balance(self, t):
            return self.work_electrical[t] == self.work_mechanical[t] * self.efficiency_motor[t]
        
    def add_isentropic_work_definition(self):
        self.delta_h_is = Expression(
            self.flowsheet().time,
            rule=lambda b, t: (
                b.properties_isentropic[t].enth_mol - b.control_volume.properties_in[t].enth_mol
            )
        )
        self.delta_h_act = Expression(
            self.flowsheet().time,
            rule=lambda b, t: (
                b.control_volume.properties_out[t].enth_mol - b.control_volume.properties_in[t].enth_mol
            )
        )
        # # Isentropic work
        # @self.Constraint(
        #     self.flowsheet().time, doc="Calculate work of isentropic process"
        # )
        # def isentropic_energy_balance(self, t):
        #     return self.work_isentropic[t] == ( self.properties_isentropic[t].enth_mol - self.control_volume.properties_in[t].enth_mol ) * self.control_volume.properties_in[t].flow_mol
            
    def initialize_build(
        blk,
        state_args=None,
        routine=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            routine : str stating which initialization routine to execute
                        * None - use routine matching thermodynamic_assumption
                        * 'isentropic' - use isentropic initialization routine
                        * 'isothermal' - use isothermal initialization routine
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        cv = blk.control_volume
        t0 = blk.flowsheet().time.first()
        state_args_out = {}

        if state_args is None:
            state_args = {}
            state_dict = cv.properties_in[t0].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        # Get initialisation guesses for outlet and isentropic states
        for k in state_args:
            if k == "pressure" and k not in state_args_out:
                # Work out how to estimate outlet pressure
                if cv.properties_out[t0].pressure.fixed:
                    # Fixed outlet pressure, use this value
                    state_args_out[k] = value(cv.properties_out[t0].pressure)
                elif blk.deltaP[t0].fixed:
                    state_args_out[k] = value(state_args[k] + blk.deltaP[t0])
                elif blk.ratioP[t0].fixed:
                    state_args_out[k] = value(state_args[k] * blk.ratioP[t0])
                else:
                    # Not obvious what to do, use inlet state
                    state_args_out[k] = state_args[k]
            elif k not in state_args_out:
                state_args_out[k] = state_args[k]

        # Initialize state blocks
        flags = cv.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
            state_args=state_args,
        )
        cv.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=False,
            state_args=state_args_out,
        )

        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------
        # Initialize Isentropic block

        blk.properties_isentropic.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )

        init_log.info_high("Initialization Step 2 Complete.")

        # Skipping step 3 because Isothermal had problems.

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 4 {}.".format(idaeslog.condition(res)))


        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.control_volume.release_state(flags, outlvl)

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{blk.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        init_log.info(f"Initialization Complete: {idaeslog.condition(res)}")

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "deltaP"):
            var_dict["Mechanical Work"] = self.work_mechanical[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Electrical Work"] = self.work_electrical[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]
        if hasattr(self, "ratioP"):
            var_dict["Pressure Ratio"] = self.ratioP[time_point]
        # if hasattr(self, "efficiency_isentropic"):
        #     var_dict["Isentropic Efficiency"] = self.efficiency_isentropic[time_point]

        return {"vars": var_dict}

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        if hasattr(self, "work_fluid"):
            for t, v in self.work_fluid.items():
                iscale.set_scaling_factor(
                    v,
                    iscale.get_scaling_factor(
                        self.control_volume.work[t], default=1, warning=True
                    ),
                )

        if hasattr(self, "work_mechanical"):
            for t, v in self.work_mechanical.items():
                iscale.set_scaling_factor(
                    v,
                    iscale.get_scaling_factor(
                        self.control_volume.work[t], default=1, warning=True
                    ),
                )

        if hasattr(self, "work_isentropic"):
            for t, v in self.work_isentropic.items():
                iscale.set_scaling_factor(
                    v,
                    iscale.get_scaling_factor(
                        self.control_volume.work[t], default=1, warning=True
                    ),
                )

        if hasattr(self, "ratioP_calculation"):
            for t, c in self.ratioP_calculation.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.control_volume.properties_in[t].pressure,
                        default=1,
                        warning=True,
                    ),
                    overwrite=False,
                )

        if hasattr(self, "fluid_work_calculation"):
            for t, c in self.fluid_work_calculation.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.control_volume.deltaP[t], default=1, warning=True
                    ),
                    overwrite=False,
                )

        if hasattr(self, "actual_work"):
            for t, c in self.actual_work.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.control_volume.work[t], default=1, warning=True
                    ),
                    overwrite=False,
                )

        if hasattr(self, "isentropic_pressure"):
            for t, c in self.isentropic_pressure.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.control_volume.properties_in[t].pressure,
                        default=1,
                        warning=True,
                    ),
                    overwrite=False,
                )

        if hasattr(self, "isentropic"):
            for t, c in self.isentropic.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.control_volume.properties_in[t].entr_mol,
                        default=1,
                        warning=True,
                    ),
                    overwrite=False,
                )

        if hasattr(self, "isentropic_energy_balance"):
            for t, c in self.isentropic_energy_balance.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.control_volume.work[t], default=1, warning=True
                    ),
                    overwrite=False,
                )

        if hasattr(self, "zero_work_equation"):
            for t, c in self.zero_work_equation.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.control_volume.work[t], default=1, warning=True
                    ),
                )

        if hasattr(self, "state_material_balances"):
            cvol = self.control_volume
            phase_list = cvol.properties_in.phase_list
            phase_component_set = cvol.properties_in.phase_component_set
            mb_type = cvol._constructed_material_balance_type
            if mb_type == MaterialBalanceType.componentPhase:
                for (t, p, j), c in self.state_material_balances.items():
                    sf = iscale.get_scaling_factor(
                        cvol.properties_in[t].get_material_flow_terms(p, j),
                        default=1,
                        warning=True,
                    )
                    iscale.constraint_scaling_transform(c, sf)
            elif mb_type == MaterialBalanceType.componentTotal:
                for (t, j), c in self.state_material_balances.items():
                    sf = iscale.min_scaling_factor(
                        [
                            cvol.properties_in[t].get_material_flow_terms(p, j)
                            for p in phase_list
                            if (p, j) in phase_component_set
                        ]
                    )
                    iscale.constraint_scaling_transform(c, sf)
            else:
                # There are some other material balance types but they create
                # constraints with different names.
                _log.warning(f"Unknown material balance type {mb_type}")



from pyomo.environ import Var, Constraint, units as pyunits, Block, Reference

def add_inverted(block: Block,component_name: str):
    """
    Adds inverted variables and constraints to a block with variable component_name.
    e.g if component_name is 'deltaP', adds deltaP_inverted variable, so we can do 
    pressure drop instead of pressure increase.
    """
    inverted_component_name = f"{component_name}_inverted"

    component = getattr(block, component_name)

    inverted_component = Var(
        component.index_set(),
        # Explanation for units:
        # Reference() turns a var/expression into a indexed object, even if it's scalar.
        # next() gets the first item from the indexed object
        # This is a way of getting around the fact that an indexed component
        # will only have its units defined at the individual index level.
        units=next(Reference(component).values()).get_units(),
    )

    block.add_component(inverted_component_name, inverted_component)

    def _inverted_rule(b, *indexes):
        return inverted_component[indexes] == -component[indexes]

    constraint = Constraint(
        component.index_set(),
        rule=_inverted_rule,
        doc=f"Inverted {component_name} Constraint",
    )
    block.add_component(f"{inverted_component_name}_constraint", constraint)

def initialise_inverted(block: Block, component_name: str):
    """
    Initialises the inverted deltaP variables to match the deltaP variable values,
    or vice versa (depending on which is fixed).

    This is generalised so you can pass any component name.
    """
    inverted_component_name = f"{component_name}_inverted"
    component = getattr(block, component_name)
    inverted_component = getattr(block, inverted_component_name)

    for indexes in component.index_set():
        if component[indexes].fixed:
            # we can savely assume that the inverted variable is not
            # fixed, otherwise this would be an overconstrained model.
            inverted_component[indexes].value = -component[indexes].value
        elif inverted_component[indexes].fixed:
            component[indexes].value = -inverted_component[indexes].value
        else:
            # neither variable is fixed,
            # so let's just leave them as is.
            # Hopefully the solver can figure it out.
            pass

def disable_inverted(block: Block, component_name: str) -> bool:
    """
    Disables the inverted variables and constraints for a given component.
    Returns True if the inverted variable was fixed, false otherwise.
    """
    inverted_component_name = f"{component_name}_inverted"
    constraint_name = f"{inverted_component_name}_constraint"

    inverted_constraint = getattr(block, constraint_name)

    inverted_constraint.deactivate()


def enable_inverted(block: Block, component_name: str):
    """
    Enables the inverted variables and constraints for a given component.
    """
    inverted_component_name = f"{component_name}_inverted"
    constraint_name = f"{inverted_component_name}_constraint"

    inverted_constraint = getattr(block, constraint_name)

    inverted_constraint.activate()


from idaes.core.base.property_base import PhysicalParameterBlock
from idaes.core import FlowsheetBlock
from property_packages.build_package import build_package

from ahuora_builder_types.flowsheet_schema import PropertyPackageType
from .flowsheet_manager_type import FlowsheetManager
from ahuora_builder_types import PropertyPackageId


def create_property_package(
    property_package_schema: PropertyPackageType, flowsheet: FlowsheetBlock
) -> PhysicalParameterBlock:
    """
    Create a property package from a schema
    """
    compounds = property_package_schema.compounds
    type = property_package_schema.type
    phases = property_package_schema.phases

    property_package = build_package(type, compounds)
    property_package_id = property_package_schema.id if property_package_schema.id != -1 else "default"

    flowsheet.add_component(f"PP_{property_package_id}", property_package)
    return property_package


class PropertyPackageManager:
    """
    Manages the property packages for a flowsheet
    """

    def __init__(self, flowsheet_manager: FlowsheetManager) -> None:
        """
        Create a new property package manager
        """
        self._flowsheet_manager = flowsheet_manager
        self._property_packages: dict[PropertyPackageId, PhysicalParameterBlock] = {}

    def load(self) -> None:
        """
        Load the property packages from the flowsheet definition
        """
        schema = self._flowsheet_manager.schema
        property_packages_schema = schema.property_packages
        for property_package_schema in property_packages_schema:
            id = property_package_schema.id
            if id in self._property_packages:
                raise Exception(f"Property package with id {id} already exists")

            property_package = create_property_package(
                property_package_schema, self._flowsheet_manager.model.fs
            )
            self._property_packages[id] = property_package

    def get(self, id: PropertyPackageId) -> PhysicalParameterBlock:
        """
        Get a property package by id
        """
        # For backwards compatibility with other tests, use id -1 as helmholtz
        fs = self._flowsheet_manager.model.fs

        if id == -1:
            if not hasattr(fs, "PP_default"):
                create_property_package(
                    PropertyPackageType(
                        id=-1,
                        type="helmholtz",
                        compounds=["h2o"],
                        phases=["Liq"]
                    ),
                    fs,
                )
            return fs.PP_default
        else:
            # get the property package by id
            if id not in self._property_packages:
                raise Exception(f"Property package with id {id} does not exist")
            return self._property_packages[id]



from typing import Any
from pyomo.environ import Expression
from pyomo.core.base.units_container import units as pyomo_units, _PyomoUnit
from sympy import Symbol
from sympy.parsing.sympy_parser import parse_expr
from pyomo.core.expr.sympy_tools import sympy2pyomo_expression, PyomoSympyBimap
from pyomo.core.base.indexed_component import IndexedComponent
from ahuora_builder.properties_manager import PropertiesManager
from .slice_manipulation import is_scalar_reference
class ExpressionParsingError(Exception):
    """Custom exception for errors during expression parsing."""
    pass

# add extra units to pyomo's units library
# could probably make this on-demand
ureg = pyomo_units._pint_registry
ureg.define("dollar = [currency]")


def handle_special_chars(expr: str) -> str:
    # replace special characters so they can be parsed
    expr = expr.replace("^", "**")
    expr = expr.replace("$", "dollar")

    return expr


def get_property_from_id(fs, property_id,time_index):
    # returns a pyomo var from a property id
    properties_map : PropertiesManager = fs.properties_map
    pyomo_object: IndexedComponent = properties_map.get_component(property_id)
    
    if pyomo_object is None:
        raise ValueError(f"Symbol with id {id} not found in model")
    # check if this is a time-indexed var, and if so get the value at the given time index
    if is_scalar_reference(pyomo_object):
        # reference with index None
        return pyomo_object[None]
    elif pyomo_object.index_set() == fs.time:
        return pyomo_object[time_index]
    else:
        raise NotImplementedError("Only 0D and 1D time-indexed properties are supported in expressions")


def evaluate_symbol(fs, symbol: str,time_index) -> Any:
    if symbol.lower() == "time" or symbol.lower() == "t":
        return float(time_index)
    if symbol.startswith("id_"):
        # get the property from flowsheet properties_map
        id = int(symbol[3:])
        return get_property_from_id(fs, id,time_index)
    else:
        # assume its a unit, eg. "m" or "kg"
        # get the unit from pint, pyomo's units library
        ureg = pyomo_units._pint_registry
        pint_unit = getattr(ureg, symbol)
        pyomo_unit = _PyomoUnit(pint_unit, ureg)
        # We want people to write expressions such as (10 * W + 5 * kW). Pyomo doesn't natively support this,
        # so we can always convert to base units.
        if symbol == "delta_degC" or symbol == "delta_degF":
            # special case, because degC is not a base unit
            return 1 * pyomo_unit
            #return _PyomoUnit(ureg.delta_degC)
        elif symbol == "degC" or symbol == "degF":
            # throw an error (we do not support this, as it is unclear what to do)
            # https://pyomo.readthedocs.io/en/6.8.1/explanation/modeling/units.html
            raise ValueError(f"Use relative temperature units (delta_degC, delta_degF) or absolute temperature units (K, degF). Cannot use {symbol} as addition and multiplication is inconsistent on non-absolute units")
        scale_factor, base_units = ureg.get_base_units(pint_unit, check_nonmult=True) # TODO: handle degC etc.
        base_pyomo_unit = _PyomoUnit(base_units, ureg)
        return pyomo_units.convert( 1 * pyomo_unit, to_units=base_pyomo_unit)


class PyomoSympyMap(PyomoSympyBimap):

    def __init__(self, model,time_index):
        self.model = model
        self.time_index = time_index

    def getPyomoSymbol(self, sympy_object: Symbol, default=None):
        if not isinstance(sympy_object, Symbol):
            return None  # It's not in pyomo, e.g a number or something
        return evaluate_symbol(self.model, sympy_object.name, self.time_index)

    def getSympySymbol(self, pyomo_object, default=None):
        raise NotImplementedError(
            "getSympySymbol not implemented, because it shouldn't be needed"
        )
        # we don't care, it only needs to go one way

    def sympyVars(self):
        raise NotImplementedError(
            "sympyVars not implemented, because it shouldn't be needed"
        )


def parse_expression(expression, model,time_index) -> Expression:
    # use the bimap to get the correct pyomo object for each symbol
    bimap = PyomoSympyMap(model,time_index)
    try:
        expression = handle_special_chars(expression)
        sympy_expr = parse_expr(expression)
        pyomo_expr = sympy2pyomo_expression(sympy_expr, bimap)
    except Exception as e:
        raise ExpressionParsingError(f"{expression}: error: {e}")
    return pyomo_expr



from .custom_heat_exchanger_1d import CustomHeatExchanger1D
from idaes.models.unit_models.heat_exchanger import HeatExchangerFlowPattern

# Map frontend string values to IDAES enums
FLOW_TYPE_MAP = {
    "countercurrent": HeatExchangerFlowPattern.countercurrent,
    "cocurrent": HeatExchangerFlowPattern.cocurrent,
}

def HeatExchanger1DWrapper(**kwargs):
    """
    Wrapper for CustomHeatExchanger1D that handles string-to-enum conversion
    and provides default values for required parameters, defaulting to collocation.
    """
    # Convert flow_type string to enum
    flow_type = kwargs.pop('flow_type', None)
    if flow_type is None:
        flow_type = "countercurrent"
    
    if isinstance(flow_type, str):
        if flow_type.lower() in FLOW_TYPE_MAP:
            flow_type_enum = FLOW_TYPE_MAP[flow_type.lower()]
        else:
            raise ValueError(f"Unknown flow_type: {flow_type}. Must be one of: {list(FLOW_TYPE_MAP.keys())}")
    else:
        flow_type_enum = flow_type
    
    kwargs['flow_type'] = flow_type_enum
    
    # Return the heat exchanger with fixed parameters
    return CustomHeatExchanger1D(**kwargs)


from .custom_valve import Valve as CustomValve, ValveFunctionType
from idaes.models.unit_models.pressure_changer import ThermodynamicAssumption
from .custom_pressure_changer import CustomPressureChanger

VALVE_FUNCTION_MAP = {
    "linear": ValveFunctionType.linear,
    "quick_opening": ValveFunctionType.quick_opening,
    "equal_percentage": ValveFunctionType.equal_percentage,
}

def ValveWrapper(**kwargs):
    enable_coefficients = kwargs.pop('enable_coefficients', False)
    valve_function = kwargs.pop('valve_function', None)

    # Set default valve function if none provided
    if valve_function is None:
        valve_function = "linear"

    # Check if valve_function is a string and map it to the corresponding callback
    if isinstance(valve_function, str):
        if valve_function in VALVE_FUNCTION_MAP:
            valve_function_callback = VALVE_FUNCTION_MAP[valve_function]
        else:
            raise ValueError(f"Unknown valve_function: {valve_function}")
    else:
        # just in case it is already a enum or a callback
        valve_function_callback = valve_function

    if enable_coefficients:
        # use the custom valve model in full
        return CustomValve(valve_function_callback=valve_function_callback, **kwargs)
    else:
        # add thermodynamic_assumption kwarg to
        # the PressureChanger model
        kwargs['thermodynamic_assumption'] = ThermodynamicAssumption.adiabatic
        kwargs['compressor'] = False
        # just use a pressure changer
        return CustomPressureChanger(**kwargs)



from idaes.core import declare_process_block_class
from idaes.models.unit_models.heater import HeaterData
from .add_initial_dynamics import add_initial_dynamics
from .inverted import add_inverted, initialise_inverted, disable_inverted, enable_inverted

@declare_process_block_class("DynamicHeater")
class DynamicHeaterData(HeaterData):
    """
    Dynamic Heater unit model class.
    This extends the Heater class to include reference variables for initial holdup and initial accumulation. 
    Which makes it easier for us to set initial conditions in the frontend.
    """

    def build(self,*args, **kwargs):
        """
        Build method for the DynamicHeaterData class.
        This method initializes the control volume and sets up the model.
        """
        super().build(*args, **kwargs)

        if hasattr(self,"deltaP"):
            # else has_pressure_change is false
            add_inverted(self, "deltaP")

        add_initial_dynamics(self)
    
    def initialize_build(
        self,*args,**kwargs,
    ):
        """
        Initialize method for the DynamicHeaterData class.
        This method initializes the control volume and sets up the model.
        """
        if hasattr(self,"deltaP"):
            initialise_inverted(self, "deltaP")
            disable_inverted(self, "deltaP")

        super().initialize_build(*args, **kwargs)

        if hasattr(self,"deltaP"):
            enable_inverted(self, "deltaP")



import pytest
import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.models.properties import iapws95
from idaes.models_extra.power_generation.unit_models.waterpipe import WaterPipe
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver

solver = get_solver()

def test_waterpipe_simple():
    '''
    Simple test for the WaterPipe unit model
    '''
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.pipe = WaterPipe(
        dynamic=False,
        property_package=m.fs.properties,
        has_holdup=True,
        has_heat_transfer=False,
        has_pressure_change=True,
        water_phase="Liq",
        contraction_expansion_at_end="None",
    )
    m.fs.pipe.diameter.fix(0.04)
    m.fs.pipe.length.fix(40)
    m.fs.pipe.number_of_pipes.fix(100)
    m.fs.pipe.elevation_change.fix(25)
    m.fs.pipe.fcorrection_dp.fix(1.0)

    state_args = {"flow_mol": 10000, "pressure": 1.3e7, "enth_mol": 18000}
    initialization_tester(m, dof=3, state_args=state_args, unit=m.fs.pipe)
    m.fs.pipe.inlet.enth_mol.fix()
    m.fs.pipe.inlet.flow_mol.fix()
    m.fs.pipe.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    assert pyo.check_optimal_termination(results)
    # Output for debugging
    print("Outlet pressure:", pyo.value(m.fs.pipe.outlet.pressure[0]))
    print("Pressure drop:", pyo.value(m.fs.pipe.deltaP[0]))
    print("Outlet enthalpy:", pyo.value(m.fs.pipe.outlet.enth_mol[0]))


from idaes.models.control.controller import PIDControllerData, ControllerType, ControllerMVBoundType, ControllerAntiwindupType, smooth_bound, smooth_heaviside
from pyomo.environ import Var
from idaes.core import UnitModelBlockData, declare_process_block_class
from pyomo.common.config import ConfigValue, In, Bool
import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
import pyomo.dae as pyodae
from idaes.core.util import scaling as iscale
import functools


@declare_process_block_class(
    "CustomVariable",
    doc="Custom variable model block. You can finally be free!",
)
class CustomVariableData(UnitModelBlockData):

    CONFIG = UnitModelBlockData.CONFIG()

    def build(self):
        """
        Build the custom variable block
        """
        super().build()
        self.variable = Var(
            self.flowsheet().time,
            initialize=0.0,
            doc="A variable you are free to do whatever you want with",
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2022
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Property package for the hydrodealkylation of toluene to form benzene
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Constraint, exp, Set, Var, Param, units as pyunits

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    ReactionParameterBlock,
    ReactionBlockDataBase,
    ReactionBlockBase,
)
from idaes.core.util.constants import Constants as const
from idaes.core.util.misc import add_object_reference

# Set up logger
_log = logging.getLogger(__name__)


@declare_process_block_class("HDAReactionParameterBlock")
class HDAReactionParameterData(ReactionParameterBlock):
    """
    Property Parameter Block Class
    Contains parameters and indexing sets associated with properties for
    superheated steam.
    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super(HDAReactionParameterData, self).build()

        self._reaction_block_class = HDAReactionBlock

        # List of valid phases in property package
        self.phase_list = Set(initialize=["Vap"])

        # Component list - a list of component identifiers
        self.component_list = Set(
            initialize=["benzene", "toluene", "hydrogen", "methane"]
        )

        # Reaction Index
        self.rate_reaction_idx = Set(initialize=["R1"])

        # Reaction Stoichiometry
        self.rate_reaction_stoichiometry = {
            ("R1", "Vap", "benzene"): 1,
            ("R1", "Vap", "toluene"): -1,
            ("R1", "Vap", "hydrogen"): -1,
            ("R1", "Vap", "methane"): 1,
            ("R1", "Liq", "benzene"): 0,
            ("R1", "Liq", "toluene"): 0,
            ("R1", "Liq", "hydrogen"): 0,
            ("R1", "Liq", "methane"): 0,
        }

        # Arrhenius Constant
        self.arrhenius = Var(
            initialize=6.3e10,
            units=pyunits.mol * pyunits.m**-3 * pyunits.s**-1 * pyunits.Pa**-1,
            doc="Arrhenius pre-exponential factor",
        )
        self.arrhenius.fix()

        # Activation Energy
        self.energy_activation = Var(
            initialize=217.6e3, units=pyunits.J / pyunits.mol, doc="Activation energy"
        )
        self.energy_activation.fix()

        # Heat of Reaction
        dh_rxn_dict = {"R1": -1.08e5}
        self.dh_rxn = Param(
            self.rate_reaction_idx,
            initialize=dh_rxn_dict,
            units=pyunits.J / pyunits.mol,
            doc="Heat of reaction",
        )

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "k_rxn": {"method": None, "units": "m^3/mol.s"},
                "reaction_rate": {"method": None, "units": "mol/m^3.s"},
            }
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


class ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """

    def initialize(blk, outlvl=0, **kwargs):
        """
        Initialization routine for reaction package.
        Keyword Arguments:
            outlvl : sets output level of initialization routine
                     * 0 = no output (default)
                     * 1 = report after each step
        Returns:
            None
        """
        if outlvl > 0:
            _log.info("{} Initialization Complete.".format(blk.name))


@declare_process_block_class("HDAReactionBlock", block_class=ReactionBlock)
class HDAReactionBlockData(ReactionBlockDataBase):
    """
    An example reaction package for saponification of ethyl acetate
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(HDAReactionBlockData, self).build()

        # Heat of reaction - no _ref as this is the actual property
        add_object_reference(self, "dh_rxn", self.config.parameters.dh_rxn)

        self.k_rxn = Var(
            initialize=0.2,
            units=pyunits.mol * pyunits.m**-3 * pyunits.s**-1 * pyunits.Pa**-1,
        )

        self.reaction_rate = Var(
            self.params.rate_reaction_idx,
            initialize=1,
            units=pyunits.mol / pyunits.m**3 / pyunits.s,
        )

        self.arrhenus_equation = Constraint(
            expr=self.k_rxn
            == self.params.arrhenius
            * exp(
                -self.params.energy_activation
                / (const.gas_constant * self.state_ref.temperature)
            )
        )

        self.rate_expression = Constraint(
            expr=self.reaction_rate["R1"]
            == self.k_rxn
            * self.state_ref.pressure
            * self.state_ref.mole_frac_phase_comp["Vap", "toluene"]
        )

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar



# Import Pyomo libraries
from stringprep import in_table_a1
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from idaes.core.util.tables import create_stream_table_dataframe

from idaes.core.util.exceptions import ConfigurationError


# Set up logger
_log = idaeslog.getLogger(__name__)


# When using this file the name "Bus" is what is imported
@declare_process_block_class("Bus")
class BusData(UnitModelBlockData):
    """
    Zero order bus model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The Bus unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The Bus unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "num_inlets",
        ConfigValue(
            default=False,
            domain=int,
            description="Number of inlets to add",
            doc="""Number of inlets to add""",
        ),
    )
    
    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)


        # Defining parameters of state block class
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet

        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        num_inlets = self.config.num_inlets

        self.inlet_list = [ "inlet_" + str(i+1) for i in range(num_inlets) ]
        

        self.inlet_blocks = []
        for name in self.inlet_list:
            # add properties_inlet_1, properties_inlet2 etc
            state_block = self.config.property_package.state_block_class(
                self.flowsheet().config.time, doc="inlet power", **tmp_dict
            )
            self.inlet_blocks.append(state_block)
            # Dynamic equivalent to self.properties_inlet_1 = stateblock
            setattr(self,"properties_" + name, state_block)
            # also add the port
            self.add_port(name=name,block=state_block)


        # Add outlet state block
        tmp_dict["defined_state"] = False  # outlet and waste block is not an inlet
        self.properties_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )

        # Add outlet port
        self.add_port(name="outlet", block=self.properties_out)


        # Add constraints
        # Usually unit models use a control volume to do the mass, energy, and momentum
        # balances, however, they will be explicitly written out in this example
        @self.Constraint(
            self.flowsheet().time,
            doc="Power usage",
        )
        def eq_power_balance(b, t):
            return (
                 sum(
                    state_block[t].power for state_block in self.inlet_blocks
                 )
                  == b.properties_out[t].power
            )


    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
    
    def initialize(blk, *args, **kwargs):
        # Just propagate the power from inlet to outlet, good simple method of initialization
        for t in blk.flowsheet().time:
            power_in = 0
            for state_block in blk.inlet_blocks:
                power_in += state_block[t].power.value
            if not blk.properties_out[t].power.fixed:
                blk.properties_out[t].power = power_in

    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
        
        io_dict = {}
        for inlet_name in self.inlet_list:
            io_dict[inlet_name] = getattr(self, inlet_name) # get a reference to the port
        
        io_dict["Outlet"] = self.outlet
        return create_stream_table_dataframe(io_dict, time_point=time_point)


import pytest
from pyomo.environ import ConcreteModel, SolverFactory, value
from idaes.core import FlowsheetBlock
from idaes.models.properties.iapws95 import Iapws95ParameterBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from ahuora_builder.custom.simple_separator import SimpleSeparator


def test_simple_separator_iapws95():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # Property package: IAPWS95 (pure water/steam)
    m.fs.properties = Iapws95ParameterBlock()

    # Add SimpleSeparator with 2 outlets
    m.fs.sep = SimpleSeparator(
        property_package=m.fs.properties,
        num_outlets=2,
    )

    # Fix inlet state
    m.fs.sep.inlet.flow_mol[0].fix(100.0)      # mol/s
    m.fs.sep.inlet.enth_mol[0].fix(50000.0)    # J/mol (example enthalpy)
    m.fs.sep.inlet.pressure[0].fix(101325.0)   # Pa

    # Fix split fractions
    m.fs.sep.outlet_1.flow_mol[0].fix(70.0)  # mol/s

    # Model should be fully specified
    assert degrees_of_freedom(m) == 0

    # Initialize and solve
    m.fs.sep.initialize()
    solver = SolverFactory("ipopt")
    results = solver.solve(m, tee=False)
    assert results.solver.termination_condition == "optimal"

    # Check material balance
    inlet_flow = value(m.fs.sep.inlet.flow_mol[0])
    out1_flow = value(m.fs.sep.outlet_1.flow_mol[0])
    out2_flow = value(m.fs.sep.outlet_2.flow_mol[0])

    assert abs(inlet_flow - (out1_flow + out2_flow)) <= 1e-6
    assert abs(out1_flow - 0.7 * inlet_flow) <= 1e-6
    assert abs(out2_flow - 0.3 * inlet_flow) <= 1e-6



import pyomo.environ as pyo
from ..infeasibilities import compute_infeasibilities, print_infeasibilities
from ...properties_manager import PropertiesManager



def test_infeasibilities(capsys):
    model = pyo.ConcreteModel()
    model.x = pyo.Var(initialize=5.0)
    model.c = pyo.Constraint(expr=model.x == 10.0)

    properties_map = PropertiesManager()
    properties_map.add(
        12345,
        pyo.Reference(model.x),
        "property name"
    )
    properties_map.add_constraint(
        12345,
        [model.c])


    print_infeasibilities(properties_map)
    captured = capsys.readouterr()
    assert "property name" in captured.out
    assert "5.0" in captured.out





from typing import Any, Callable
from abc import ABC, abstractmethod
from idaes.core.base.property_base import PhysicalParameterBlock
from idaes.core.base.reaction_base import ReactionParameterBlock
from ..flowsheet_manager_type import FlowsheetManager
from ahuora_builder_types import PropertyPackageId




class AdapterBase(ABC):
    """
    Base class for adapter methods
    """

    @abstractmethod
    def run(self, value, flowsheet_manager: FlowsheetManager) -> Any:
        pass


class PropertyPackage(AdapterBase):
    """Get a property package by ID"""

    def run(self, id: PropertyPackageId, flowsheet_manager: FlowsheetManager) -> PhysicalParameterBlock:
        return flowsheet_manager.property_packages.get(id)

class ReactionPackage(AdapterBase):
    """Get a reaction package"""
    def run(self, id: str, flowsheet_manager: FlowsheetManager) -> ReactionParameterBlock:
        return flowsheet_manager.model.fs.reaction_params
    
class ReactorPropertyPackage(AdapterBase):
     def run(self,value,flowsheet_manager) -> PhysicalParameterBlock:
        return flowsheet_manager.model.fs.peng_robinson   


class Value(AdapterBase):
    """Do not modify the value passed"""

    def run(self, value: Any,flowsheet_manager) -> Any:
        return value


class Constant(AdapterBase):
    """Always return the argument passed"""
    def __init__(self,constant) -> None:
        self._constant = constant

    def run(self, value, flowsheet_manager) -> Callable:
        return self._constant


class Dictionary(AdapterBase):
    """Validate a dictionary against a schema and adapt the values"""
    def __init__(self,schema) -> None:
        self._schema = schema

    def run(self, dictionary: dict, flowsheet_manager) -> Callable[[dict[str, Callable]], dict]:
        schema = self._schema

        result = {}
        for item, adapter in schema.items():
            # TODO: do we need to handle optional items better?
            result[item] = adapter.run(dictionary.get(item, None),flowsheet_manager)
        return result


class PowerPropertyPackage(AdapterBase):
    def run(self,value,flowsheet_manager) -> PhysicalParameterBlock:
        return flowsheet_manager.model.fs.power_pp # See __init__ method of Flowsheet_Manager.py 

class acPropertyPackage(AdapterBase):
    def run(self,value,flowsheet_manager) -> PhysicalParameterBlock:
        return flowsheet_manager.model.fs.ac_pp # See __init__ method of Flowsheet_Manager.py 

class TransformerPropertyPackage(AdapterBase):
    def run(self,value,flowsheet_manager) -> PhysicalParameterBlock:
        return flowsheet_manager.model.fs.tr_pp # See __init__ method of Flowsheet_Manager.py 





# Import Pyomo libraries
from stringprep import in_table_a1
from pyomo import environ
from pyomo.environ import (
    Constraint,
    Set,
    Var,
    Suffix,
    units as pyunits,
    atan
)
from pyomo.environ import Reals
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.contrib.fbbt import interval
from idaes.core.util.model_statistics import degrees_of_freedom
# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    MaterialBalanceType,
    MaterialFlowBasis,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.models.unit_models import separator
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError, BurntToast, PropertyNotSupportedError

from idaes.models.unit_models import separator
#Import enum
from enum import Enum

# Set up logger
_log = idaeslog.getLogger(__name__)

import math
# Enumerate options for balances
class SplittingType(Enum):
    """
    Enum of supported material split types.
    """

    totalFlow = 1
    phaseFlow = 2
    componentFlow = 3
    phaseComponentFlow = 4



# When using this file the name "transmissionLine" is what is imported
@declare_process_block_class("transmissionLine")
class transmissionLineData(UnitModelBlockData):
    """
    Zero order transmission line model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The Bus unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The Bus unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "num_inlets",
        ConfigValue(
            default=False,
            domain=int,
            description="Number of inlets to add",
            doc="Number of inlets to add",
        ),
    )
    CONFIG.declare(
        "num_outlets",
        ConfigValue(
            default=False,
            domain=int,
            description="Number of outlets to add",
            doc="Number of outlets to add",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
            **default** - MaterialBalanceType.useDefault.
            **Valid values:** {
            **MaterialBalanceType.useDefault - refer to property package for default
            balance type
            **MaterialBalanceType.none** - exclude material balances,
            **MaterialBalanceType.componentPhase** - use phase component balances,
            **MaterialBalanceType.componentTotal** - use total component balances,
            **MaterialBalanceType.elementTotal** - use total element balances,
            **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )


    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package

                # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        # Add inlet block
        # tmp_dict = dict(**self.config.property_package_args)
        # tmp_dict["parameters"] = self.config.property_package
        # tmp_dict["defined_state"] = True  # inlet block is an inlet
        # self.properties_in = self.config.property_package.state_block_class(
        #     self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        # )
       
        # Add outlet and waste block
        tmp_dict["defined_state"] = False  # outlet and waste block is not an inlet
        self.properties_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )
         # Add second outlet block
        tmp_dict["defined_state"] = False 
        self.properties_out2 = self.config.property_package.state_block_class(
            self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        )

        # Add ports - oftentimes users interact with these rather than the state blocks
        # self.add_port(name="inlet_1", block=self.properties_in)
        # self.add_port(name="inlet_2", block=self.properties_in2)
        self.add_port(name="outlet_1", block=self.properties_out)
        self.add_port(name="outlet_2", block=self.properties_out2)       
       
        self.power_transfer = Var(self.flowsheet().config.time,
            initialize=1.0,
            doc="Power transferred between busses",
            units = pyunits.W

        )

        # Add constraints:
        @self.Constraint(
            self.flowsheet().time,
            doc = "Power balance"
        )
        def eq_power_balance(b,t):
            return self.properties_out2[t].power == self.power_transfer[t]
        
        @self.Constraint(
            self.flowsheet().time,
            doc = "Power calculation"
        )
        def eq_power_calc(b,t):
            return self.properties_out[t].power == self.power_transfer[t] * -1


                
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
    
    def initialize(blk, *args, **kwargs):
       for i in blk.properties_out.index_set():
            if not blk.properties_out[i].power.fixed:
                blk.properties_out[i].power = blk.power_transfer[i].value   


    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
       

   



# Pyomo core
import pyomo.environ as pyo
from pyomo.environ import (
    Constraint,
    Expression,
    Param,
    PositiveReals,
    RangeSet,
    Suffix,
    Var,
    value,
    units as UNIT,
)
from pyomo.core.base.reference import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool

# IDAES core
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    StateBlock,
)
from idaes.core.util import scaling
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.math import smooth_min, smooth_max
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.solvers import get_solver
from idaes.core.initialization import ModularInitializerBase
from idaes.core.util.model_statistics import degrees_of_freedom

# Logger
import idaes.logger as idaeslog

# Typing
from typing import List


__author__ = "Ahuora Centre for Smart Energy Systems, University of Waikato, New Zealand"

# Set up logger
_log = idaeslog.getLogger(__name__)

class SimpleHeaderInitializer(ModularInitializerBase):
    """Initialize a Header unit block with staged seeding and solves.

    This routine performs a two-stage initialization:
    1) Seed inlet and internal state variables, relax selected constraints, and
       perform a first solve.
    2) Reactivate/tighten constraints and perform a second solve.

    Args:
        blk: The Header unit model block to initialize.
        **kwargs: Optional keyword arguments:
            solver: A Pyomo/IDAES solver object. If not provided, uses ``get_solver()``.
            solver_options (dict): Options to set on the solver, e.g. tolerances.
            outlvl: IDAES log level (e.g., ``idaeslog.WARNING``).

    Returns:
        pyomo.opt.results.results_.SolverResults: The result object from the final solve.

    Notes:
        - Inlet state blocks are initialized via their own ``initialize`` if available.
        - Mixed state is seeded from inlet totals/minimums (pressure) and average
          enthalpy; works with temperature- or enthalpy-based property packages.
        - Temporary seeds/relaxations are undone, leaving original DOF intact.
    """

    def initialize(self, blk, **kwargs):
        # --- Solver setup
        solver = kwargs.get("solver", None) or get_solver()
        solver_options = kwargs.get("solver_options", {})
        for k, v in solver_options.items():
            solver.options[k] = v

        outlvl = kwargs.get("outlvl", idaeslog.WARNING)
        log = idaeslog.getLogger(__name__)

        # --- Time index
        t0 = blk.flowsheet().time.first()

        # --- 1) Initialize inlet state blocks
        inlet_blocks = list(blk.inlet_blocks)
        if len(inlet_blocks) < 1:
            raise ValueError("No inlet added to header.")
        
        for sb in inlet_blocks:
            if hasattr(sb, "initialize"):
                sb.initialize(outlvl=outlvl)

        # --- 2) Aggregate inlet info for seeding mixed state block
        F_mixed = sum(
            value(sb[t0].flow_mol) 
            for sb in inlet_blocks
        )
        P_mixed = min(
            value(sb[t0].pressure) 
            for sb in inlet_blocks
        )
        E_mixed = sum(
            value(sb[t0].flow_mol * sb[t0].enth_mol, 0.0) 
            for sb in inlet_blocks
        )
        if F_mixed > 0:
            h_mixed = E_mixed / F_mixed
        else:
            # Seed from the first inlet’s enthalpy (no double subscripting)
            first_inlet = inlet_blocks[0]
            h_mixed = value(first_inlet[t0].enth_mol)

        # --- 3) Seed mixed_state: flow, pressure, enthalpy
        ms = blk.mixed_state
        ms[t0].flow_mol.set_value(
            F_mixed
        )
        ms[t0].pressure.set_value(
            P_mixed
        )
        ms[t0].enth_mol.set_value(
            h_mixed
        ) 
        ms.initialize(outlvl=outlvl)     

        # --- 4) Seed outlet_states with pressure, enthalpy
        flow_undefined = []
        defined_flow = 0
        for sb in blk.outlet_blocks:
            sb[t0].pressure.set_value(
                value(ms[t0].pressure)
            )
            sb[t0].enth_mol.set_value(
                value(ms[t0].enth_mol)
            )
            if sb in [blk.outlet_condensate_state, blk.outlet_vent_state]:
                sb[t0].flow_mol.set_value(
                    0.0
                )
            else:
                if value(sb[t0].flow_mol, exception=False) is None:
                    flow_undefined.append(sb) 
                else:
                    defined_flow += value(sb[t0].flow_mol)
        
        tot_undefined_flow = max(sum(value(sb[t0].flow_mol) for sb in blk.inlet_blocks) - defined_flow, 0)
        for sb in flow_undefined:
            sb[t0].flow_mol.set_value(
                tot_undefined_flow / len(flow_undefined)
            )

        for sb in blk.outlet_blocks:        
            sb.initialize(outlvl=outlvl)

        res2 = solver.solve(blk, tee=False)
        log.info(f"Header init status: {res2.solver.termination_condition}")

        return res2

def _make_config_block(config):
    """Declare configuration options for the Header unit.

    Declares property package references and integer counts for inlets and outlets.

    Args:
        config (ConfigBlock): The mutable configuration block to populate.

    Raises:
        ValueError: If invalid option values are provided by the caller (via IDAES).
    """

    config.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
        ),
    )
    config.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
        ),
    )
    config.declare(
        "num_inlets",
        ConfigValue(
            default=1,
            domain=In(list(range(1, 100))),
            description="Number of utility providers at inlets.",
        ),
    )
    config.declare(
        "num_outlets",
        ConfigValue(
            default=1,
            domain=In(list(range(0, 100))),
            description="Number of utility users at outlets." \
            "Excludes outlets associated with condensate and vent flows.",
        ),
    )
    config.declare(
        "is_liquid_header",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Flag for selecting liquid or vapour (including steam and other gases).",
        ),
    )
@declare_process_block_class("simple_header")
class SimpleHeaderData(UnitModelBlockData):
    """Thermal utility header unit operation.

    The Header aggregates multiple inlet providers and distributes utility to
    multiple users, with optional venting, condensate removal (or liquid overflow), heat loss, and
    pressure loss. A mixed (intermediate) state is used for balances and
    pressure/enthalpy coupling across outlets.

    Key features:
        - Material, energy, and momentum balances with smooth min/max functions.
        - Vapour/liquid equilibrium calculation for mixed state.
        - Shared mixed enthalpy across outlets of the same phase.
        - Computed excess flow from an overall flow balance.
        - Optional heat and pressure losses.

    Attributes:
        inlet_list (list[str]): Names for inlet ports.
        outlet_list (list[str]): Names for outlet ports (incl. condensate/ and vent).
        inlet_blocks (list): StateBlocks for all inlets.
        outlet_blocks (list): StateBlocks for all outlets.
        mixed_state: Intermediate mixture StateBlock.
        heat_loss (Var): Heat loss from the header (W).
        pressure_loss (Var): Pressure drop from inlet minimum to mixed state (Pa).
        makeup_flow_mol (Var): Required inlet makeup molar flow (mol/s).
    """

    default_initializer=SimpleHeaderInitializer
    CONFIG = UnitModelBlockData.CONFIG()
    _make_config_block(CONFIG)

    def build(self) -> None:
        # 1. Inherit standard UnitModelBlockData properties and functions
        super().build()

        # 2. Validate input parameters are valid
        self._validate_model_config()

        # 3. Create lists of ports with state blocks to add
        self.inlet_list = self._create_inlet_port_name_list()
        self.outlet_list = self._create_outlet_port_name_list()

        # 4. Declare ports, state blocks and state property bounds 
        self.inlet_blocks = self._add_ports_with_state_blocks(
            stream_list=self.inlet_list,
            is_inlet=True,
            has_phase_equilibrium=False,
            is_defined_state=True,
        )
        self.outlet_blocks = self._add_ports_with_state_blocks(
            stream_list=self.outlet_list,
            is_inlet=False,
            has_phase_equilibrium=False,
            is_defined_state=False
        )
        self._internal_blocks = self._add_internal_state_blocks()
        self._add_bounds_to_state_properties()
        self._outlet_supply_blocks = self._create_custom_state_lists()
        
        # 4. Declare references, variables and expressions for external and internal use
        self._create_references()
        self._create_variables()
        self._create_expressions()

        # 5. Set balance equations
        self._add_material_balances()
        self._add_energy_balances()
        self._add_momentum_balances()
        self._add_additional_constraints()

        # 6. Other
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)
        self.split_flow = self._create_flow_map_references()

    def _validate_model_config(self) -> bool:
        """Validate configuration for inlet and outlet counts.

        Raises:
            ValueError: If ``num_inlets < 1`` or ``num_outlets < 1``.
        """
        if self.config.num_inlets < 1:
            raise ValueError("Header requires at least one provider (num_inlets >= 1).")
        if self.config.num_outlets < 1:
            raise ValueError("Header requires at least one user (num_outlets >= 1).")
        return True

    def _create_inlet_port_name_list(self) -> List[str]:
        """Build ordered inlet port names.

        Returns:
            list[str]: Names ``["inlet_1", ..., "inlet_N"]`` based on ``num_inlets``.
        """
        return [
            f"inlet_{i+1}" for i in range(self.config.num_inlets)
        ]

    def _create_outlet_port_name_list(self) -> List[str]:
        """Build ordered outlet port names.

        Returns:
            list[str]: Names ``["outlet_1", ..., "outlet_n", "outlet_condensate", "outlet_vent"]``.
        """        
        return [
            f"outlet_{i+1}" for i in range(self.config.num_outlets)
        ] + ["outlet_condensate"] + ["outlet_vent"]

    def _add_ports_with_state_blocks(self, 
                                     stream_list: List[str], 
                                     is_inlet: List[str], 
                                     has_phase_equilibrium: bool=False,
                                     is_defined_state: bool=None,
                                     ) -> List[StateBlock]:
        """Construct StateBlocks and expose them as ports.

        Creates a StateBlock per named stream and attaches a corresponding inlet or
        outlet Port. Inlet blocks are defined states; outlet blocks are calculated states.

        Args:
            stream_list (list[str]): Port/StateBlock base names to create.
            is_inlet (bool): If True, create inlet ports with ``defined_state=True``;
                otherwise create outlet ports with ``defined_state=False``.
            has_phase_equilibrium (bool)

        Returns:
            list: The created StateBlocks, in the same order as ``stream_list``.
        """
        # Create empty list to hold StateBlocks for return
        state_block_ls = []

        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = has_phase_equilibrium
        if is_defined_state == None:
            tmp_dict["defined_state"] = True if is_inlet else False
        else:
            tmp_dict["defined_state"] = is_defined_state

        # Create an instance of StateBlock for all streams
        for s in stream_list:
            sb = self.config.property_package.build_state_block(
                self.flowsheet().time, doc=f"Thermophysical properties at {s}", **tmp_dict
            )
            setattr(
                self, s + "_state", 
                sb
            )
            state_block_ls.append(sb)
            add_fn = self.add_inlet_port if is_inlet else self.add_outlet_port
            add_fn(
                name=s,
                block=sb,
            )

        return state_block_ls

    def _add_internal_state_blocks(self) -> List[StateBlock]:
        """Create the intermediate (mixed) StateBlock.

        The mixed state:
            - Has phase equilibrium enabled.
            - Is not a defined state (solved from balances).
        """
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = True
        tmp_dict["defined_state"] = False

        self.mixed_state = self.config.property_package.build_state_block(
            self.flowsheet().time, 
            doc=f"Thermophysical properties at intermediate mixed state.", 
            **tmp_dict
        )
        return [
            self.mixed_state
        ]

    def _add_bounds_to_state_properties(self) -> None:
        """Add lower and/or upper bounds to state properties.

        - Set nonnegativity lower bounds on all inlet/outlet molar flows.
        """
        for sb in (self.inlet_blocks + self.outlet_blocks):
            for t in sb:
                sb[t].flow_mol.setlb(0.0)     
    
    def _create_custom_state_lists(self) -> List[StateBlock]:
        """Partition outlet names into vapour outlets and capture their StateBlocks.

        Populates:
            - ``_outlet_supply_list``: Outlet names excluding condensate and vent.
            - ``_outlet_supply_blocks``: Corresponding StateBlocks.
        """        
        self._outlet_supply_list = [
            v for v in self.outlet_list 
            if not v in ["outlet_condensate", "outlet_vent"]
        ]
        return [
            getattr(self, n + "_state") 
            for n in self._outlet_supply_list
        ]

    def _create_references(self) -> None:
        """Create convenient References.

        Creates references to mixed_state properties:
            - ``total_flow_mol`` 
            - ``total_flow_mass``
            - ``pressure`` 
            - ``temperature`` 
            - ``enth_mol`` 
            - ``enth_mass`` 
            - ``vapor_frac``
        """
        self.total_flow_mol = Reference(
            self.mixed_state[:].flow_mol
        )
        self.total_flow_mass = Reference(
            self.mixed_state[:].flow_mass
        )
        self.pressure = Reference(
            self.mixed_state[:].pressure
            )
        self.temperature = Reference(
            self.mixed_state[:].temperature
        )
        self.enth_mol = Reference(
            self.mixed_state[:].enth_mol
            )
        self.enth_mass = Reference(
            self.mixed_state[:].enth_mass
            )
        self.vapor_frac = Reference(
            self.mixed_state[:].vapor_frac
        )

    def _create_variables(self) -> None:
        """Create required variables.

        Creates:
            - ``heat_loss`` (W)
            - ``pressure_loss`` (Pa)
        """
        self.heat_loss = Var(
            self.flowsheet().time, 
            initialize=0.0, 
            doc="Heat loss",
            units=UNIT.W
        )
        self.pressure_loss = Var(
            self.flowsheet().time, 
            initialize=0.0, 
            doc="Pressure loss",
            units=UNIT.Pa
        )                

    def _create_expressions(self) -> None:
        """Create convenient Expressions.

        Creates:
            - ``balance_flow_mol`` (mol/s)
            - ``degree_of_superheat`` (K)
            - ``makeup_flow_mol`` (mol/s)
            - ``_partial_total_flow_mol`` (mol/s): used for scaling purposes in a material balance
        """
        self.degree_of_superheat = Expression(
            self.flowsheet().time,
            rule=lambda b, t: b.temperature[t] - b.outlet_condensate_state[t].temperature
        )
        self._partial_total_flow_mol = Expression(
            self.flowsheet().time,
            rule=lambda b, t: (
                sum(
                    o[t].flow_mol
                    for o in (b.inlet_blocks + b._outlet_supply_blocks)
                )                
            )
        )       
        self.balance_flow_mol = Expression(
            self.flowsheet().time,
            rule=lambda b, t: (
                sum(
                    i[t].flow_mol
                    for i in b.inlet_blocks
                )
                -
                sum(
                    o[t].flow_mol
                    for o in (
                        b._outlet_supply_blocks + 
                        [
                            b.outlet_vent_state 
                            if self.config.is_liquid_header 
                            else b.outlet_condensate_state
                        ]
                    )
                )                
            )
        )        
        self.makeup_flow_mol = Expression(
            self.flowsheet().time,
            rule=lambda b, t: (
                (
                    b.outlet_condensate_state[t].flow_mol 
                    if self.config.is_liquid_header 
                    else b.outlet_vent_state[t].flow_mol
                ) 
                - 
                b.balance_flow_mol[t]
            )
        )

    def _add_material_balances(self) -> None:
        """Material balance equations summary.

        Introduces:
            - ``_partial_total_flow_mol``: Sum of known inlet and vapour outlet flows,
            used for scaling a smooth vent calculation.

        Constraints:
            - ``mixed_state_material_balance``: Mixed flow equals total inlet flow.
            - ``vent_flow_balance``: Depends on the header's primary phase: liquid vs gas
                                        If gas header, smoothly enforces nonnegative vent flow.
                                        If liquid header, determines flow from mixed-state vapour fraction
            - ``condensate_flow_balance``: Depends on the header's primary phase: liquid vs gas
                                        If gas header, determines flow from mixed-state vapour fraction
                                        If liquid header, smoothly enforces nonnegative condensate flow.
        """

        @self.Constraint(
                self.flowsheet().time, 
                doc="Mixed state material balance",
                )
        def mixed_state_material_balance(b, t):
            return (
                b.mixed_state[t].flow_mol
                == 
                sum(
                    i[t].flow_mol
                    for i in b.inlet_blocks
                )
            )

        eps = 1e-5  # smoothing parameter; smaller = closer to exact max, larger = smoother
        if self.config.is_liquid_header:
            # Assigns excess liquid flow to outlet_condensate
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Condensate flow balance." \
                    "Determines the positive amount of excess flow that exits through outlet_condensate"
                    )
            def condensate_flow_balance(b, t):
                return (
                    b.outlet_condensate_state[t].flow_mol
                    == 
                    smooth_max(
                        b.balance_flow_mol[t] / (b._partial_total_flow_mol[t] + 1e-6),
                        0.0,
                        eps,
                    ) * (b._partial_total_flow_mol[t] + 1e-6)
                )
                        
            # Removes any gas/vapour from a liquid header
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Vent balance."
                    )
            def vent_flow_balance(b, t):
                return b.outlet_vent_state[t].flow_mol == (
                    b.mixed_state[t].flow_mol * b.mixed_state[t].vapor_frac
                )
        else:
            # Assigns excess steam/vapour flow to outlet_vent
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Vent flow balance." \
                    "Determines the positive amount of excess flow that exits through the vent"
                    )
            def vent_flow_balance(b, t):
                return (
                    b.outlet_vent_state[t].flow_mol
                    == 
                    smooth_max(
                        b.balance_flow_mol[t] / (b._partial_total_flow_mol[t] + 1e-6),
                        0.0,
                        eps,
                    ) * (b._partial_total_flow_mol[t] + 1e-6)
                )

            # Removes any condensate/liquid from a steam/gas header
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Condensate balance."
                    )
            def condensate_flow_balance(b, t):
                return (
                    b.outlet_condensate_state[t].flow_mol 
                    ==
                    b.mixed_state[t].flow_mol * (1 - b.mixed_state[t].vapor_frac)
                )
            
    def _add_energy_balances(self) -> None:
        """Energy balance equations summary.

        Introduces:
            - ``_liq_out_enth_mol``: Shared molar enthalpy for all liquid outlets,
            including the condensate.
            - ``_vap_out_enth_mol``: Shared molar enthalpy for all vapour outlets,
            including the vent.
            
        Constraints:
            - ``inlets_to_mixed_state_energy_balance``: Inlet energy to mixed state (+ heat loss).
            - ``mixed_state_to_outlets_energy_balance``: Mixed state to all outlets.
            - ``molar_enthalpy_equality_eqn``: Common vapour enthalpy across vapour outlets and vent.
        """
        @self.Constraint(self.flowsheet().time, doc="Inlets to mixed state energy balance including heat loss")
        def inlets_to_mixed_state_energy_balance(b, t):
            return (
                b.mixed_state[t].flow_mol * b.mixed_state[t].enth_mol 
                + b.heat_loss[t]
                == 
                sum(
                    i[t].flow_mol * i[t].enth_mol
                    for i in b.inlet_blocks
                )
            )
        @self.Constraint(
                self.flowsheet().time, 
                doc="Mixed state to outlets energy balance"
                )
        def mixed_state_to_outlets_energy_balance(b, t):
            return (
                b.mixed_state[t].enth_mol 
                * 
                sum(
                    o[t].flow_mol
                    for o in b.outlet_blocks
                )
                == 
                sum(
                    o[t].flow_mol * o[t].enth_mol
                    for o in b.outlet_blocks
                )           
            )   
        if self.config.is_liquid_header:
            self._liq_out_enth_mol = Var(
                self.flowsheet().time, 
                initialize=42.0 * 18, 
                doc="Molar enthalpy of the liquid outlets", 
                units=UNIT.J / UNIT.mol
            )
            @self.Constraint(
                self.flowsheet().time,
                self._outlet_supply_blocks + [self.outlet_condensate_state], # exclude vent outlet
                doc="All liquid outlets (incl. condensate) share a common liquid enthalpy",
                )
            def molar_enthalpy_equality_eqn(b, t, o):
                return (
                    o[t].enth_mol 
                    == 
                    b._liq_out_enth_mol[t]
                )
        else:
            self._vap_out_enth_mol = Var(
                self.flowsheet().time, 
                initialize=2700.0 * 18, 
                doc="Molar enthalpy of the vapour outlets", 
                units=UNIT.J / UNIT.mol
            )        
            @self.Constraint(
                self.flowsheet().time,
                self._outlet_supply_blocks + [self.outlet_vent_state], # exclude condensate outlet
                doc="All vapour outlets (incl. vent) share a common vapour enthalpy",
                )
            def molar_enthalpy_equality_eqn(b, t, o):
                return (
                    o[t].enth_mol 
                    == 
                    b._vap_out_enth_mol[t]
                )

    def _add_momentum_balances(self) -> None:
        """Momentum balance equations summary.

        Computes the minimum inlet pressure via a sequential smooth minimum and
        sets the mixed-state pressure to that minimum minus ``pressure_loss``,
        then enforces equality to every outlet pressure.

        Notes:
            - Uses IDAES ``smooth_min`` for differentiable minimum pressure.
            - ``_eps_pressure`` is a smoothing parameter (units of pressure).
        """
        inlet_idx = RangeSet(len(self.inlet_blocks))
        # Get units metadata
        units = self.mixed_state.params.get_metadata()
        # Add variables
        self._minimum_pressure = Var(
            self.flowsheet().time,
            inlet_idx,
            doc="Variable for calculating minimum inlet pressure",
            units=units.get_derived_units("pressure"),
        )
        self._eps_pressure = Param(
            mutable=True,
            initialize=1e-3,
            domain=PositiveReals,
            doc="Smoothing term for minimum inlet pressure",
            units=units.get_derived_units("pressure"),
        )
        # Calculate minimum inlet pressure
        @self.Constraint(
            self.flowsheet().time,
            inlet_idx,
            doc="Calculation for minimum inlet pressure",
        )
        def minimum_pressure_constraint(b, t, i):
            if i == inlet_idx.first():
                return (
                    b._minimum_pressure[t, i] 
                    == 
                    (b.inlet_blocks[i - 1][t].pressure)
                )
            else:
                return (
                    b._minimum_pressure[t, i] 
                    ==
                    smooth_min(
                        b._minimum_pressure[t, i - 1],
                        b.inlet_blocks[i - 1][t].pressure,
                        b._eps_pressure,
                    )
                )
        # Set mixed pressure to minimum inlet pressure minus any pressure loss
        @self.Constraint(
            self.flowsheet().time,
            doc="Pressure equality constraint from minimum inlet to mixed state",
        )
        def mixture_pressure(b, t):
            return (
                b.mixed_state[t].pressure 
                == 
                b._minimum_pressure[t, inlet_idx.last()] - b.pressure_loss[t]
            )
        # Set outlet pressures to mixed pressure
        @self.Constraint(
            self.flowsheet().time,
            self.outlet_blocks,
            doc="Pressure equality constraint from mixed state to outlets",
        )
        def pressure_equality_eqn(b, t, o):
            return (
                b.mixed_state[t].pressure 
                == 
                o[t].pressure
            )

    def _add_additional_constraints(self) -> None:
        """Add auxiliary constraints and bounds.
        
        - Fix vent vapour fraction to near one (near 100% vapour).
        OR
        - Fix condensate vapour fraction to a small value (near 100% liquid).
        """
        if self.config.is_liquid_header:
            @self.Constraint(self.flowsheet().time, doc="Vent vapour fraction.")
            def vent_vapour_fraction(b, t):
                return (
                    b.outlet_vent_state[t].vapor_frac 
                    == 
                    1 #1 - 1e-6
                )
        else:
            @self.Constraint(self.flowsheet().time, doc="Condensate vapour fraction.")
            def condensate_vapour_fraction(b, t):
                return (
                    b.outlet_condensate_state[t].vapor_frac 
                    == 
                    0 #1e-6
                )

    def _create_flow_map_references(self):
        """Create a two-key Reference for outlet flows over time and outlet name.

        Builds a mapping ``(t, outlet_name) -> outlet_state[t].flow_mol`` and exposes it
        as a Reference for compact access to outlet flow splits.

        Returns:
            pyomo.core.base.reference.Reference: A Reference indexed by ``(time, outlet)``.
        """
        self.outlet_idx = pyo.Set(initialize=self.outlet_list)
        # Map each (t, o) to the outlet state's flow var
        ref_map = {}
        for o in self.outlet_list:
            if o != "vent":
                outlet_state_block = getattr(self, f"{o}_state")
                for t in self.flowsheet().time:
                    ref_map[(t, o)] = outlet_state_block[t].flow_mol

        return Reference(ref_map)

    def calculate_scaling_factors(self):
        """Assign scaling factors to improve numerical conditioning.

        Sets scaling factors for performance and auxiliary variables. If present,
        also scales the shared vapour enthalpy variable ``_vap_out_enth_mol``.
        """
        super().calculate_scaling_factors()
        scaling.set_scaling_factor(self.heat_loss, 1e-6)
        scaling.set_scaling_factor(self.pressure_loss, 1e-6)
        scaling.set_scaling_factor(self.balance_flow_mol, 1e-3)
        scaling.set_scaling_factor(self._partial_total_flow_mol, 1e-3)
        if hasattr(self, "_vap_out_enth_mol"):
            scaling.set_scaling_factor(self._vap_out_enth_mol, 1e-6)

    def _get_stream_table_contents(self, time_point=0):
        """Create a stream table for all inlets and outlets.

        Args:
            time_point (int | float): Time index at which to extract stream data.

        Returns:
            pandas.DataFrame: A tabular view suitable for reporting via
            ``create_stream_table_dataframe``.
        """
        io_dict = {}

        for inlet_name in self.inlet_list:
            io_dict[inlet_name] = getattr(self, inlet_name)

        for outlet_name in self.outlet_list:
            io_dict[outlet_name] = getattr(self, outlet_name)

        return create_stream_table_dataframe(io_dict, time_point=time_point)

    def _get_performance_contents(self, time_point=0, is_full_report=True):
        """Collect performance results for reporting.

        Args:
            time_point (int | float): Time index at which to report values.
            is_full_report (bool): Flag for full or partial performance report.

        Returns:
            dict: A report of internal unit model results.
        """
        return (
            {
                "vars": {
                    "Heat Loss": self.heat_loss[time_point],
                    "Pressure Drop": self.pressure_loss[time_point],
                    "Mass Flow": self.mixed_state[time_point].flow_mass,
                    "Molar Flow": self.mixed_state[time_point].flow_mol,
                    "Balance Flow": self.balance_flow_mol[time_point],
                    "Pressure": self.mixed_state[time_point].pressure,
                    "Temperature": self.mixed_state[time_point].temperature,
                    "Degree of Superheat": self.degree_of_superheat[time_point],
                    "Vapour Fraction": self.mixed_state[time_point].vapor_frac,
                    "Mass Specific Enthalpy": self.mixed_state[time_point].enth_mass,
                    "Molar Specific Enthalpy": self.mixed_state[time_point].enth_mol,
                }
            } if is_full_report else {
                "vars": {
                    "Balance Flow": self.balance_flow_mol[time_point],
                    "Pressure": self.mixed_state[time_point].pressure,
                    "Temperature": self.mixed_state[time_point].temperature,
                    "Degree of Superheat": self.degree_of_superheat[time_point],
                }
            }

        )
            
    def initialize(self, *args, **kwargs):
        """Initialize the Header unit using :class:`SimpleHeaderInitializer`.

        Args:
            *args: Forwarded to ``SimpleHeaderInitializer.initialize``.
            **kwargs: Forwarded to ``SimpleHeaderInitializer.initialize`` (e.g., solver, options).

        Returns:
            pyomo.opt.results.results_.SolverResults: Results from the initializer's solve.
        """
        init = SimpleHeaderInitializer()
        return init.initialize(self, *args, **kwargs)



from idaes.models.unit_models import (
    Pump,
    Heater,
    Mixer,
    StateJunction,
    HeatExchanger,
    Turbine,
    PressureChanger,
)
from idaes.models.unit_models.heat_exchanger import delta_temperature_underwood_callback
from idaes.models.unit_models.pressure_changer import ThermodynamicAssumption
from idaes.models.unit_models.separator import SplittingType, EnergySplittingType
from ahuora_builder.custom.heat_exchanger_1d_wrapper import HeatExchanger1DWrapper
from . import adapter_methods as methods
from ..custom.custom_heat_exchanger import CustomHeatExchanger
from ..custom.PySMOModel import PySMOModel
from ..custom.energy.link import Link
from ..custom.energy.power_property_package import PowerParameterBlock
from ..custom.energy.ac_property_package import acParameterBlock
from ..custom.energy.transformer_property_package import transformerParameterBlock
from ..custom.updated_pressure_changer import Compressor
from ..custom.energy.solar import Solar
from ..custom.custom_valve import Valve
from ..custom.energy.wind import Wind
from ..custom.energy.energy_mixer import EnergyMixer
from ..custom.energy.bus    import Bus
from ..custom.energy.acBus    import acBus
from ..custom.energy.transmissionLine import transmissionLine
from ..custom.valve_wrapper import ValveWrapper
from ..custom.energy.load import Load
from ..custom.energy.hydro import Hydro
from ..custom.energy.transformer import Transformer
from ..custom.energy.storage import Storage
from ..custom.direct_steam_injection import Dsi
from ..custom.translator import GenericTranslator
from idaes.models.control.controller import  ControllerType, ControllerMVBoundType
from ..custom.PIDController import PIDController2
from ..custom.energy.grid import Grid
from ..custom.custom_heater import DynamicHeater
from ..custom.custom_tank import CustomTank
from ..custom.custom_compressor import CustomCompressor
from ..custom.custom_pump import CustomPump
from ..custom.integration_block import IntegrationBlock
from idaes.models.unit_models.cstr import CSTR
from ..custom.energy.energy_splitter import EnergySplitter
from ..custom.custom_turbine import CustomTurbine
from ..custom.energy.mainDistributionBoard import MDB
from ..custom.custom_separator import CustomSeparator
from ..custom.custom_variable import CustomVariable
from ..custom.thermal_utility_systems.steam_header import SteamHeader
from ..custom.thermal_utility_systems.header import simple_header
from ..custom.reactions.hda_stoich import HDAStoichiometricReactor
from ..custom.custom_heat_exchanger_1d import CustomHeatExchanger1D
from ..custom.thermal_utility_systems.simple_heat_pump import SimpleHeatPump
from ..custom.thermal_utility_systems.steam_user import SteamUser
from ..custom.thermal_utility_systems.desuperheater import Desuperheater
from idaes.models_extra.power_generation.unit_models.waterpipe import WaterPipe
from ..custom.SimpleEffectivenessHX_DH import HeatExchangerEffectiveness
from idaes.models.unit_models.heat_exchanger_lc import HeatExchangerLumpedCapacitance
from idaes.models_extra.column_models.plate_heat_exchanger import PlateHeatExchanger
from ..custom.custom_cooler import CustomCooler

class UnitModelConstructor:
    """
    Schema for adapter library items
    """

    def __init__(self, model_constructor: type, arg_parsers: dict):
        # idaes model constructor
        self.model_constructor: type = model_constructor
        # dictionary of argument parsers
        self.arg_parsers: dict[str,methods.AdapterBase] = arg_parsers



"""
A dictionary of adapters for the idaes model constructors
"""
AdapterLibrary: dict[str, UnitModelConstructor] = {
        "pump": UnitModelConstructor(
            CustomPump,
            {
                "property_package": methods.PropertyPackage(),
                "dynamic": methods.Constant(False),
                "power_property_package": methods.PowerPropertyPackage(),

            },
        ),
        "heater": UnitModelConstructor(
            DynamicHeater,
            {
                "property_package": methods.PropertyPackage(),
                "has_pressure_change": methods.Value(),
                "dynamic": methods.Value(),
                "has_holdup": methods.Value(),
            },
        ),
        "compressor": UnitModelConstructor(
            CustomCompressor,
            {
                "property_package": methods.PropertyPackage(),
                "power_property_package": methods.PowerPropertyPackage(),
                "dynamic": methods.Constant(False),
            },
        ),
        "turbine": UnitModelConstructor(
            CustomTurbine,
            {
                "property_package": methods.PropertyPackage(),
                "power_property_package": methods.PowerPropertyPackage(),
                "dynamic": methods.Constant(False),
                "calculation_method": methods.Constant("isentropic"),
            },
        ),
        "willans_turbine": UnitModelConstructor(
            CustomTurbine,
            {
                "property_package": methods.PropertyPackage(),
                "power_property_package": methods.PowerPropertyPackage(),
                "dynamic": methods.Constant(False),
                "calculation_method": methods.Constant("simple_willans"),
            },
        ),
        "pl_turbine": UnitModelConstructor(
            CustomTurbine,
            {
            "property_package": methods.PropertyPackage(),
            "power_property_package": methods.PowerPropertyPackage(),
            "dynamic": methods.Constant(False),
            "calculation_method": methods.Constant("part_load_willans"),  
            }
        ),
        "dts_turbine": UnitModelConstructor(
            CustomTurbine,
            {
            "property_package": methods.PropertyPackage(),
            "power_property_package": methods.PowerPropertyPackage(),
            "dynamic": methods.Constant(False),
            "calculation_method": methods.Constant("Tsat_willans"),  
            }
        ),
        "bs_turbine": UnitModelConstructor(
            CustomTurbine,
            {
            "property_package": methods.PropertyPackage(),
            "power_property_package": methods.PowerPropertyPackage(),
            "dynamic": methods.Constant(False),
            "calculation_method": methods.Constant("BPST_willans"),  
            }
        ),
        "cs_turbine": UnitModelConstructor(
            CustomTurbine,
            {
            "property_package": methods.PropertyPackage(),
            "power_property_package": methods.PowerPropertyPackage(),
            "dynamic": methods.Constant(False),
            "calculation_method": methods.Constant("CT_willans"),  
            }
         ),  
        "mixer": UnitModelConstructor(
            Mixer,
            {
                "property_package": methods.PropertyPackage(),
                "num_inlets": methods.Value(),
            },
        ),
        "Tank": UnitModelConstructor(
            CustomTank,
            {
                "property_package": methods.PropertyPackage(),
                "dynamic": methods.Value(),
                "has_holdup": methods.Value(),
                "tank_type": methods.Constant("vertical_cylindrical_tank"),
                "has_heat_transfer": methods.Constant(True),
            }
        ),
        "heatExchanger": UnitModelConstructor(
            CustomHeatExchanger,
            {
                "delta_temperature_callback": methods.Constant(
                    delta_temperature_underwood_callback
                ),
                "hot_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(True),
                    }
                ),
                "cold_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(True),
                    }
                ),
                "dynamic": methods.Constant(False),
            },
        ),
        "heatPump": UnitModelConstructor(
            SimpleHeatPump,
            {
                "source": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(False),
                    }
                ),
                "sink": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(False),
                    }
                ),
                "dynamic": methods.Constant(False),
            },
        ),
        "valve": UnitModelConstructor(
            ValveWrapper, # type: ignore
            {
                "property_package": methods.PropertyPackage(),
                "enable_coefficients": methods.Value(),
                "valve_function": methods.Value(),
                "dynamic": methods.Constant(False),
            },
        ),
        "cooler": UnitModelConstructor(
            CustomCooler,
            {
                "property_package": methods.PropertyPackage(),
                "has_pressure_change": methods.Value(),
                "dynamic": methods.Constant(False),
            },
        ),
        "splitter": UnitModelConstructor(
            CustomSeparator,
            {
                "property_package": methods.PropertyPackage(),
                "num_outlets": methods.Value(),
            },
        ),
        "phaseSeparator": UnitModelConstructor(
            CustomSeparator,
            {
                "property_package": methods.PropertyPackage(),
                "num_outlets": methods.Value(),
                "split_basis": methods.Constant(
                    SplittingType.phaseFlow
                ),
                "energy_split_basis": methods.Constant(
                    EnergySplittingType.enthalpy_split
                ),
            },
        ),
        "compoundSeparator": UnitModelConstructor(
            CustomSeparator,
            {
                "property_package": methods.PropertyPackage(),
                "num_outlets": methods.Value(),
                "split_basis": methods.Constant(
                    SplittingType.componentFlow
                ),
            },
        ),
        "machineLearningBlock": UnitModelConstructor(PySMOModel, {
            "property_package": methods.PropertyPackage(),
            "model": methods.Value(),
            "ids": methods.Value(),
            "unitopNames": methods.Value(),
            "num_inlets": methods.Value(),
            "num_outlets": methods.Value(),
        }),
        "link": UnitModelConstructor(
            Link,
            {
                "property_package": methods.PowerPropertyPackage(),
            },
        ),
        "solar": UnitModelConstructor(
            Solar,
            {
                "property_package": methods.PowerPropertyPackage(),
            },
        ),
        "wind": UnitModelConstructor(
            Wind,
            {
                "property_package": methods.PowerPropertyPackage(),
            },
        ),
        "energy_mixer": UnitModelConstructor(
            EnergyMixer,
            {
                "property_package": methods.PowerPropertyPackage(),
                "num_inlets": methods.Value(),
            },
        ),
        "bus": UnitModelConstructor(
            Bus,
            {
                "property_package": methods.PowerPropertyPackage(),
                "num_inlets": methods.Value(),
                
            },
         ), 
        "acBus": UnitModelConstructor(
            acBus,
            {
                "property_package": methods.acPropertyPackage(),
                "num_inlets": methods.Value(),
                "num_outlets": methods.Value(),
            },       
        ),
        "transmissionLine": UnitModelConstructor(
            transmissionLine,
            {
                "property_package": methods.PowerPropertyPackage(),
                "num_inlets": methods.Value(),
                "num_outlets": methods.Value(),
            },
        ),
        "load": UnitModelConstructor(
            Load,
            {
                "property_package": methods.PowerPropertyPackage(),
            },
        ),
        "hydro": UnitModelConstructor(
            Hydro,
            {
                "property_package": methods.PowerPropertyPackage(),
            },
        ),
    
        "transformer": UnitModelConstructor(
            Transformer,
            {
                "property_package": methods.PowerPropertyPackage(),
            },
        ),
        "storage": UnitModelConstructor(
            Storage,
            {
                "property_package": methods.PowerPropertyPackage(),
                "has_holdup": methods.Value(),
                "dynamic": methods.Value(),
            },
        ),
        "direct_steam_injection": UnitModelConstructor(
            Dsi,
            {
                "property_package": methods.PropertyPackage(),
                "steam_property_package": methods.PropertyPackage(),
            },
        ),
        "translator": UnitModelConstructor(
            GenericTranslator,
            {
                "inlet_property_package": methods.PropertyPackage(),
                "outlet_property_package": methods.PropertyPackage(),
                "translator_type": methods.Value(),
            },
        ),
        "pid_controller": UnitModelConstructor(
            PIDController2,
            {
                "dynamic": methods.Constant(True),
                "controller_type": methods.Constant(ControllerType.PI),
                "calculate_initial_integral": methods.Constant(True),
                "mv_bound_type": methods.Constant(ControllerMVBoundType.SMOOTH_BOUND),
            },
        ),
        "custom_variable": UnitModelConstructor(
            CustomVariable,
            {},
        ),
        "grid": UnitModelConstructor(
            Grid,
            {
                "property_package": methods.PowerPropertyPackage(),
            },
        ),
        "integration": UnitModelConstructor(
            IntegrationBlock,
            {
            },
        ),
        "stoich_hda": UnitModelConstructor(
            HDAStoichiometricReactor,
            {
                "property_package": methods.ReactorPropertyPackage(),
                "has_heat_transfer": methods.Constant(True),
                #"has_heat_of_reaction": methods.Value(),
                #"has_pressure_change": methods.Value(),
                "has_holdup": methods.Constant(False),
                "dynamic": methods.Constant(False),
                "reaction_package": methods.ReactionPackage(),
            }, 
        ),
        "stoich_hda": UnitModelConstructor(
            HDAStoichiometricReactor,
            {
                "property_package": methods.ReactorPropertyPackage(),
                "has_heat_transfer": methods.Constant(True),
                #"has_heat_of_reaction": methods.Value(),
                #"has_pressure_change": methods.Value(),
                "has_holdup": methods.Constant(False),
                "dynamic": methods.Constant(False),
                "reaction_package": methods.ReactionPackage(),
            }, 
        ),
        "RCT_CSTR": UnitModelConstructor(
            CSTR,
            {
                "property_package": methods.ReactorPropertyPackage(),
                "has_heat_transfer": methods.Value(),
                "has_heat_of_reaction": methods.Value(),
                "has_pressure_change": methods.Value(),
                "has_holdup": methods.Value(),
                "dynamic": methods.Value(),
                "reaction_package": methods.ReactionPackage(),
            },
            
        ),
        "energy_splitter": UnitModelConstructor(
            EnergySplitter,
            {
                "property_package": methods.PowerPropertyPackage(),
                "num_inlets": methods.Value(),
                "num_outlets": methods.Value(),
            },
        ),
        "mdb": UnitModelConstructor(
            MDB,
            {
                "property_package": methods.PowerPropertyPackage(),
                "num_inlets": methods.Value(),
                "num_outlets": methods.Value(),
            }
        ),
        "header": UnitModelConstructor(
            SteamHeader,
            {
                "property_package": methods.PropertyPackage(),
                "num_inlets": methods.Value(),
                "num_outlets": methods.Value(),
            },
        ),
        "simple_header": UnitModelConstructor(
            simple_header,
            {
                "property_package": methods.PropertyPackage(),
                "num_inlets": methods.Value(),
                "num_outlets": methods.Value(),
            },
        ),
        "heat_exchanger_1d": UnitModelConstructor(
            HeatExchanger1DWrapper,
            {
                "hot_side": methods.Dictionary({
                    "property_package": methods.PropertyPackage(),
                    "transformation_method": methods.Value(),
                    "transformation_scheme": methods.Value(),
                    "has_pressure_change": methods.Constant(False),
                }),
                "cold_side": methods.Dictionary({
                    "property_package": methods.PropertyPackage(),
                    "transformation_method": methods.Value(),
                    "transformation_scheme": methods.Value(),
                    "has_pressure_change": methods.Constant(False),
                }),
                "finite_elements": methods.Value(),
                "collocation_points": methods.Value(),
                "flow_type": methods.Value(),
            },
        ),
        "steam_user": UnitModelConstructor(
            SteamUser,
            {
                "property_package": methods.PropertyPackage(),
            }
        ),
         "desuperheater": UnitModelConstructor(
            Desuperheater,
            {
                "property_package": methods.PropertyPackage(),
            }
        ),
        "heat_exchanger_ntu": UnitModelConstructor(
            HeatExchangerEffectiveness,
            {
                "hot_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(True),
                    }
                ),
                "cold_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(True),
                    }
                ),
                "dynamic": methods.Constant(False),
            },
        ),
        "waterpipe": UnitModelConstructor(
            WaterPipe,
            {
                "property_package": methods.PropertyPackage(),
                "has_heat_transfer": methods.Constant(False),
                "has_pressure_change": methods.Constant(True),
            }
        ),
        "heat_exchanger_lc": UnitModelConstructor(
            HeatExchangerLumpedCapacitance,
            {
                "hot_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(False),
                    }
                ),
                "cold_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(False),
                    }
                ),
                "dynamic_heat_balance": methods.Constant(False),
            },
        ),
        "plate_heat_exchanger": UnitModelConstructor(
            PlateHeatExchanger,
            {
                "hot_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(True),
                    }
                ),
                "cold_side": methods.Dictionary(
                    {
                        "property_package": methods.PropertyPackage(),
                        "has_pressure_change": methods.Constant(True),
                    }
                ),
                "dynamic": methods.Constant(False),
            },
        ),
    }



import pytest

import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from ahuora_builder.custom.thermal_utility_systems.desuperheater import Desuperheater
from property_packages.build_package import build_package

eps_rel = 1e-4 
eps_abs = 1e-2

def desuperheater_base_case(deltaT_superheat, T_water, T_steam, P_steam, heat_loss, pressure_loss):
    # This defines the base case for all tests
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.water = build_package(
        "helmholtz", 
        ["water"], 
        ["Liq", "Vap"]
    )

    # Create SteamUser
    m.fs.desuperheater = Desuperheater(
        property_package=m.fs.water
    )

    # Set inlet conditions for steam inlet
    m.fs.desuperheater.inlet_steam.flow_mol.fix(1000)    # mol/s
    m.fs.desuperheater.inlet_steam.pressure.fix(P_steam)  # Pa
    m.fs.desuperheater.inlet_steam.enth_mol.fix(
        m.fs.water.htpx((T_steam + 273.15) * pyo.units.K, 
        m.fs.desuperheater.inlet_steam.pressure[0]),
    )  # J/mol

    # Define the state variables
    m.fs.desuperheater.deltaT_superheat.fix(
        deltaT_superheat * pyo.units.K
    ) # K
    m.fs.desuperheater.bfw_temperature[:].fix(
        (T_water + 273.15) * pyo.units.K
    ) # K
    m.fs.desuperheater.heat_loss[:].fix(heat_loss)  # W -> heat loss
    m.fs.desuperheater.pressure_loss[:].fix(pressure_loss)  # Pa -> pressure loss

    # Check degrees of freedom
    assert degrees_of_freedom(m.fs) == 0
    
    # Initialize the model
    m.fs.desuperheater.initialize()
    
    # Solve the model
    opt = pyo.SolverFactory("ipopt_v2")
    results = opt.solve(m, tee=False)

    return m, results
    
def test_desuperheater_with_zero_superheat_in_outlet_flow():
    P = 400*1000
    T = 200
    T_w = 80
    dT_sup = 0
    Q_loss = 0
    dP = 0
    
    m, results = desuperheater_base_case(
        deltaT_superheat=dT_sup, 
        T_water=T_w, 
        T_steam=T, 
        heat_loss=Q_loss, 
        pressure_loss=dP, 
        P_steam=P,
    )

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert (pyo.value(m.fs.desuperheater.outlet_state[0].temperature) - pyo.value(m.fs.desuperheater._int_sat_vap_state[0].temperature)) == pytest.approx(dT_sup, rel=eps_rel, abs=eps_abs)
    assert pyo.value(m.fs.desuperheater.outlet_state[0].pressure) == pytest.approx(P - dP, rel=eps_rel, abs=eps_abs)

def test_desuperheater_with_zero_superheat_in_outlet_flow_and_pressure_loss():
    # Case where there is zero liquid flow and there is more known inlet than outlet flows
    P = 400*1000
    T = 200
    T_w = 80
    dT_sup = 0
    Q_loss = 0
    dP = 1000
    
    m, results = desuperheater_base_case(
        deltaT_superheat=dT_sup, 
        T_water=T_w, 
        T_steam=T, 
        heat_loss=Q_loss, 
        pressure_loss=dP, 
        P_steam=P,
    )
    
    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert (pyo.value(m.fs.desuperheater.outlet_state[0].temperature) - pyo.value(m.fs.desuperheater._int_sat_vap_state[0].temperature)) == pytest.approx(dT_sup, rel=eps_rel, abs=eps_abs)
    assert pyo.value(m.fs.desuperheater.outlet_state[0].pressure) == pytest.approx(P - dP, rel=eps_rel, abs=eps_abs)

def test_desuperheater_with_10_superheat_in_outlet_flow():
    # Case where there is zero liquid flow and there is more known inlet than outlet flows
    P = 400*1000
    T = 200
    T_w = 80
    dT_sup = 10
    Q_loss = 0
    dP = 0
    
    m, results = desuperheater_base_case(
        deltaT_superheat=dT_sup, 
        T_water=T_w, 
        T_steam=T, 
        heat_loss=Q_loss, 
        pressure_loss=dP, 
        P_steam=P,
    )

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert (pyo.value(m.fs.desuperheater.outlet_state[0].temperature) - pyo.value(m.fs.desuperheater._int_sat_vap_state[0].temperature)) == pytest.approx(dT_sup, rel=eps_rel, abs=eps_abs)
    assert pyo.value(m.fs.desuperheater.outlet_state[0].pressure) == pytest.approx(P - dP, rel=eps_rel, abs=eps_abs)

def test_desuperheater_with_5_superheat_in_outlet_flow_plus_losses():
    # Case where there is zero liquid flow and there is more known inlet than outlet flows
    P = 400*1000
    T = 200
    T_w = 80
    dT_sup = 5
    Q_loss = 10000
    dP = 5000
    
    m, results = desuperheater_base_case(
        deltaT_superheat=dT_sup, 
        T_water=T_w, 
        T_steam=T, 
        heat_loss=Q_loss, 
        pressure_loss=dP, 
        P_steam=P,
    )
    
    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert (pyo.value(m.fs.desuperheater.outlet_state[0].temperature) - pyo.value(m.fs.desuperheater._int_sat_vap_state[0].temperature)) == pytest.approx(dT_sup, rel=eps_rel, abs=eps_abs)
    assert pyo.value(m.fs.desuperheater.outlet_state[0].pressure) == pytest.approx(P - dP, rel=eps_rel, abs=eps_abs)



from pyomo.environ import Var, Constraint, Reference, units as pyunits
from pyomo.common.numeric_types import value
from pyomo.opt.results.solver import check_optimal_termination
from idaes.core import declare_process_block_class
from idaes.core.util import scaling as iscale
from idaes.models.unit_models.heat_exchanger_1D import HeatExchanger1DData, HX1DInitializer
from idaes.models.unit_models.heat_exchanger import HeatExchangerFlowPattern
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError
import idaes.logger as idaeslog

class CustomHX1DInitializer(HX1DInitializer):
    """
    Use our custom control-volume initialize (no source port-member fixing).
    """
    def initialize_control_volume(self, cv, state_args=None):
        return initialize(
            cv,
            state_args=state_args,
            outlvl=self.get_output_level(),
        )

@declare_process_block_class(
    "CustomHeatExchanger1D",
    doc="1D Heat Exchanger with overall U tied to local heat_transfer_coefficient.",
)
class CustomHeatExchanger1DData(HeatExchanger1DData):
    # Use our initializer so both sides use the custom CV initialize
    default_initializer = CustomHX1DInitializer

    CONFIG = HeatExchanger1DData.CONFIG()

    def build(self):
        super().build()
        # Ends of the tube along the length axis (start and end positions)
        x_first = self.hot_side.length_domain.first()
        x_last = self.hot_side.length_domain.last()

        # Hot side: inlet at start, outlet at end
        x_hot_in, x_hot_out = x_first, x_last

        # Cold side: depends on flow pattern
        if self.config.flow_type == HeatExchangerFlowPattern.cocurrent:
            x_cold_in, x_cold_out = x_first, x_last
        else:
            x_cold_in, x_cold_out = x_last, x_first

        # Time-only inlet/outlet views of the boundary states (no extra vars/cons)
        self.hot_side.properties_in = Reference(self.hot_side.properties[:, x_hot_in])
        self.hot_side.properties_out = Reference(self.hot_side.properties[:, x_hot_out])
        self.cold_side.properties_in = Reference(self.cold_side.properties[:, x_cold_in])
        self.cold_side.properties_out = Reference(self.cold_side.properties[:, x_cold_out])

        # Overall U
        self.overall_heat_transfer_coefficient = Var(
            self.flowsheet().time,
            initialize=500.0,
            bounds=(1.0, 1e5),
            units=pyunits.W / pyunits.m**2 / pyunits.K,
            doc="Overall (constant along length) heat transfer coefficient U.",
        )

        @self.Constraint(self.flowsheet().time, self.hot_side.length_domain)
        def overall_heat_transfer_coefficient_def(b, t, x):
            return b.overall_heat_transfer_coefficient[t] == b.heat_transfer_coefficient[t, x]

        iscale.set_scaling_factor(self.overall_heat_transfer_coefficient, 1e-3)

    def initialize_build(
        self,
        hot_side_state_args=None,
        cold_side_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        duty=None,
    ):
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        opt = get_solver(solver, optarg)

        # Sync length values
        if self.length.fixed:
            self.cold_side.length.set_value(self.length)
        elif self.cold_side.length.fixed:
            self.length.set_value(self.cold_side.length)

        # Initialize control volumes with length fixed
        Lfix = self.hot_side.length.fixed
        self.hot_side.length.fix()
        flags_hot_side = initialize(
            self.hot_side,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=hot_side_state_args,
        )
        if not Lfix:
            self.hot_side.length.unfix()

        Lfix = self.cold_side.length.fixed
        self.cold_side.length.fix()
        # Use our custom CV initialize here as well
        flags_cold_side = initialize(
            self.cold_side,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=cold_side_state_args,
        )
        if not Lfix:
            self.cold_side.length.unfix()

        init_log.info_high("Initialization Step 1 Complete.")

        # Fixed-duty solve
        hot_units = self.hot_side.config.property_package.get_metadata().get_derived_units
        cold_units = self.cold_side.config.property_package.get_metadata().get_derived_units
        t0 = self.flowsheet().time.first()

        # Use inlet indices for each side
        x_hot_in = self.hot_side.length_domain.first()
        x_cold_in = self.cold_side.length_domain.first() if self.config.flow_type == HeatExchangerFlowPattern.cocurrent else self.cold_side.length_domain.last()

        if duty is None:
            duty = value(
                0.25
                * self.heat_transfer_coefficient[t0, x_hot_in]
                * self.area
                * (
                    self.hot_side.properties[t0, x_hot_in].temperature
                    - pyunits.convert(
                        self.cold_side.properties[t0, x_cold_in].temperature,
                        to_units=hot_units("temperature"),
                    )
                )
            )
        else:
            duty = pyunits.convert_value(duty[0], from_units=duty[1], to_units=hot_units("power"))

        duty_per_length = value(duty / self.length)

        # Fix heat duties
        for v in self.hot_side.heat.values():
            v.fix(-duty_per_length)
        for v in self.cold_side.heat.values():
            v.fix(pyunits.convert_value(duty_per_length, to_units=cold_units("power")/cold_units("length"), from_units=hot_units("power")/hot_units("length")))

        # Deactivate heat duty constraints and solve
        self.heat_transfer_eq.deactivate()
        self.heat_conservation.deactivate()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Unfix heat duties and re-activate constraints
        for v in self.hot_side.heat.values():
            v.unfix()
        for v in self.cold_side.heat.values():
            v.unfix()
        self.heat_transfer_eq.activate()
        self.heat_conservation.activate()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        release_state(self.hot_side, flags_hot_side)
        release_state(self.cold_side, flags_cold_side)

        if res is not None and not check_optimal_termination(res):
            raise InitializationError(f"{self.name} failed to initialize successfully. See logs.")

        init_log.info("Initialization Complete.")


def initialize(
        blk,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        optarg=None,
        solver=None,
        hold_state=True,
    ):
        """
        Initialization routine for 1D control volume.

        Keyword Arguments:
            state_args: a dict of arguments to be passed to the property
                package(s) to provide an initial state for initialization
                (see documentation of the specific property package) (default = {}).
            outlvl: sets output level of initialization routine
            optarg: solver options dictionary object (default=None, use
                default solver options)
            solver: str indicating which solver to use during initialization
                (default = None)
            hold_state: flag indicating whether the initialization routine
                should unfix any state variables fixed during initialization,
                (default = True). **Valid values:**
                **True** - states variables are not unfixed, and a dict of
                returned containing flags for which states were fixed
                during initialization, **False** - state variables are
                unfixed after initialization by calling the release_state
                method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization else the release state is
            triggered.
        """
        if optarg is None:
            optarg = {}

        # Get inlet state if not provided
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="control_volume")

        # Provide guesses if none
        if state_args is None:
            blk.estimate_states(always_estimate=True)


        if state_args is None:
            # If no initial guesses provided, estimate values for states
            blk.estimate_states(always_estimate=True)

        # Initialize state blocks
        flags = blk.properties.initialize(
            state_args=state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        try:
            # TODO: setting state_vars_fixed may not work for heterogeneous
            # systems where a second control volume is involved, as we cannot
            # assume those state vars are also fixed. For now, heterogeneous
            # reactions should ignore the state_vars_fixed argument and always
            # check their state_vars.
            blk.reactions.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_vars_fixed=True,
            )
        except AttributeError:
            pass

        init_log.info("Initialization Complete")

        # Unfix state variables except for source block
        blk.properties.release_state(flags)

        return {}

def release_state(blk, flags, outlvl=idaeslog.NOTSET):
    # No-op: nothing was fixed at the CV level in our custom initialize
    return


from idaes.models.unit_models.pressure_changer import PressureChangerData
from .inverted import add_inverted, initialise_inverted
from idaes.core import declare_process_block_class

@declare_process_block_class("CustomPressureChanger")
class CustomPressureChangerData(PressureChangerData):
    """
    Custom Pressure Changer model that includes inverted deltaP property.
    """

    def build(self, *args, **kwargs):
        """
        Build method for the CustomPressureChangerData class.
        This method initializes the control volume and sets up the model.
        """
        super().build(*args, **kwargs)

        # add deltaP_inverted as a property
        add_inverted(self, "deltaP")

    def initialize_build(
        self,*args,**kwargs,
    ):
        """
        Initialization method for the CustomPressureChangerData class.

        Args:
            state_args (dict): Arguments to be passed to the state block
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """
        initialise_inverted(self, "deltaP")

        super().initialize_build(*args, **kwargs)



from idaes.core import declare_process_block_class
from idaes.models.unit_models.stoichiometric_reactor import StoichiometricReactorData
from pyomo.environ import Var, Reference

@declare_process_block_class("HDAStoichiometricReactor")
class HDAStoichiometricReactorData(StoichiometricReactorData):
    def build(self,*args, **kwargs):
        super().build(*args, **kwargs)
        # reaction id in https://github.com/IDAES/examples/blob/50065c4cc4de96a8dc1cad833101df2d96574ac4/idaes_examples/mod/hda/hda_reaction.py#L4
        self.hda_extent = Reference(self.rate_reaction_extent[:,"R1"])


from typing import Optional
from pydantic import BaseModel
from ahuora_builder_types import FlowsheetSchema
from ahuora_builder_types.flowsheet_schema import SolvedFlowsheetSchema
from ahuora_builder_types.payloads.solve_request_schema import IdaesSolveRequestPayload
from .flowsheet_manager import FlowsheetManager


class SolveModelResult(BaseModel):
    input_flowsheet: FlowsheetSchema
    output_flowsheet: SolvedFlowsheetSchema
    solve_index: Optional[int] = None
    scenario_id: Optional[int] = None
    task_id: int
    timing: dict

def solve_model(solve_request: IdaesSolveRequestPayload) -> SolveModelResult:
    """Solves the model and returns the results"""
    flowsheet = FlowsheetManager(solve_request.flowsheet)
    #print(solve_request.flowsheet.model_dump_json())
    flowsheet.load()
    flowsheet.initialise()
    flowsheet.report_statistics()
    if solve_request.perform_diagnostics:
        flowsheet.diagnose_problems()
    flowsheet.check_model_valid()
    flowsheet.solve()
    flowsheet.optimize()
    result = flowsheet.serialise()

    return SolveModelResult(
        input_flowsheet=solve_request.flowsheet,
        output_flowsheet=result,
        solve_index=solve_request.solve_index,
        scenario_id=solve_request.scenario_id,
        task_id=solve_request.task_id,
        timing=flowsheet.timing.close()
    )


from typing import Any
from idaes.core import FlowsheetBlock

from ahuora_builder_types.flowsheet_schema import PropertyPackageType
from ahuora_builder_types.payloads import BuildStateRequestSchema
from ahuora_builder_types.unit_model_schema import SolvedPropertyValueSchema
from .property_package_manager import create_property_package
from .methods.adapter import fix_block, serialize_properties_map, deactivate_fixed_guesses
from .methods.BlockContext import BlockContext
from pyomo.environ import ConcreteModel
from .properties_manager import PropertiesManager
from pyomo.environ import Block, assert_optimal_termination, SolverFactory
from pyomo.core.base.constraint import ScalarConstraint
from .flowsheet_manager import build_flowsheet
from idaes.core.util.model_statistics import degrees_of_freedom, number_unused_variables, number_activated_equalities, number_unfixed_variables, number_unfixed_variables_in_activated_equalities
from idaes.core.util.model_diagnostics import DiagnosticsToolbox

def solve_state_block(schema: BuildStateRequestSchema) -> list[SolvedPropertyValueSchema]:
    m, sb = build_state(schema.property_package)
    # if there's only one compound in the property package, remove the mole_frac_comp (as it's over specified)
    if len(schema.property_package.compounds) == 1:
        del schema.properties["mole_frac_comp"]
    
    block_ctx = BlockContext(m.fs)
    fix_block(sb, schema.properties, m.fs, block_ctx)

    # If the degrees of freedom are not zero, don't try solve.
    # However, the degree of freedom logic ignores any variables that aren't actually used. So if temperature 
    # and pressure are both not specified, and there are no constraints for them either, it decides that
    # they are not part of the solution, and says there's 0 degrees of freedom.
    # so instead, we actually check there are unfixed variables that are not in activated equalities.
    # TODO: Check why adding and degrees_of_freedom(sb) == 0 makes pr fail (python manage.py test core.auxiliary.tests.test_Compounds)
    if number_unfixed_variables(sb) - number_unfixed_variables_in_activated_equalities(sb) != 0:
        return [] # This means no properties are returned, so the backend won't update anything.
    block_ctx.apply_elimination()
    
    # initialise the state block, which will perform a solve
    sb.initialize(outlvl=1)
    deactivate_fixed_guesses(m.fs.guess_vars)

    return serialize_properties_map(m.fs)


def build_state(schema: PropertyPackageType) -> Any:  # PropertyPackageSchema
    m = build_flowsheet(dynamic=False)
    # create the property package and state block
    property_package = create_property_package(schema, m)
    state_block = property_package.build_state_block(m.fs.time, defined_state=True)
    m.fs.add_component(f"PP_{schema.id}_state", state_block)

    return m, state_block


def get_state_vars(schema: PropertyPackageType) -> Any:
    _, state_block = build_state(schema)

    return state_block[0].define_state_vars()



from pyomo.environ import value
from pyomo.network import Arc
from idaes.core.util.tables import _get_state_from_port
from pyomo.core.base.expression import ScalarExpression
from ahuora_builder_types.arc_schema import TearGuessSchema
from .flowsheet_manager_type import FlowsheetManager
from ahuora_builder_types import PortId
from .methods.adapter import fix_var


class TearManager:
    """
    Manages the tears in the flowsheet
    """

    def __init__(self, flowsheet_manager: FlowsheetManager):
        """
        Create a new tear manager
        """
        self._flowsheet_manager = flowsheet_manager
        self._tears: list[Arc] = []

    def load(self):
        """
        Load all the tears (from the recycle unitops)
        """
        schema = self._flowsheet_manager.schema

        for arc_schema in schema.arcs:
            if arc_schema.tear_guess:
                self.add_tear(arc_schema)

    def add_tear(self, arc_schema):
        """
        Add a tear to the flowsheet
        """
        portId = arc_schema.destination
        guess = arc_schema.tear_guess

        port = self._flowsheet_manager.ports.get_port(portId)
        arc = port.arcs()[0]
        self._tears.append(arc)

        """
        During model loading, we add in all the constraints and fix the variables
        for this port. Since it is a tear, things need to be deactivated/unfixed
        to ensure 0 degrees of freedom

        If the state block is defined by constraints, we need to solve it to get the
        correct values for the state variables. Then we deactivate the constraints
        since we will be fixing the state variables instead (where applicable).
        """
        
        # hardcoding time indexes to [0] for now
        time_indexes = [0]

        # need to get the correct value for state variables before running
        # sequential decomposition (since all state variables are fixed
        # during sequential decomposition).
        sb = _get_state_from_port(port, time_indexes[0])
        # deactivate any guesses
        for key, value in guess.items():
            var = getattr(sb, key)
            if value != True:
                if isinstance(var,ScalarExpression):
                    pass # we want to solve with constraints
                else:
                    # this might give too few dof if we have other constraints. We need to unfix
                    # this value
                    var.unfix()

        if len(list(sb.constraints.component_objects())) > 0:
            # state block is defined by some constraints, so we need to solve it.
            # initialize the state block, which should put the correct value in
            # the state variables
            sb.parent_component().initialize()
            # deactivate the state block constraints, since we should use the
            # state variables as guesses (or fix them instead)
            sb.constraints.deactivate()

        blk = sb.parent_component()
        for key in sb.define_state_vars():
            if guess.get(key, False):
                # deactivate the equality constraint (expanded arcs)
                expanded_arc = getattr(
                    self._flowsheet_manager.model.fs, arc._name + "_expanded"
                )
                equality_constraint = getattr(expanded_arc, key + "_equality")
                equality_constraint.deactivate()

                # fix the variable
                for b in blk.values():
                    getattr(b, key).fix()

            else:
                # unfix this variable
                for b in blk.values():
                    getattr(b, key).unfix()



from typing import Any

from pyomo.network import Port
from pyomo.environ import value as pyo_value, Var
from pyomo.core.base.constraint import ScalarConstraint
from pyomo.core.base.units_container import units
from idaes.core import UnitModelBlock
from idaes.core.util.tables import _get_state_from_port
from idaes.core.util.model_serializer import StoreSpec, from_json

from ahuora_builder.methods.BlockContext import BlockContext
from .adapter_library import UnitModelConstructor
from ..flowsheet_manager_type import FlowsheetManager
from ahuora_builder_types import  UnitModelSchema
from .adapter import fix_block
from ahuora_builder.methods.adapter import get_component

def add_unit_model_to_flowsheet(
    unit_model_def: UnitModelSchema,
    adapter_constructor: UnitModelConstructor,
    flowsheet_manager: FlowsheetManager,
) -> UnitModelBlock: 
    """
    Add the unit model to the flowsheet.
    Args:
        unit_model_def: The definition of the unit model to be added to the flowsheet.
        adapter_schema: Methods used to create the model and parse the arguments.
        flowsheet_manager: Flowsheet manager to interact with the flowsheet.
    """
    # Create the model
    idaes_model: UnitModelBlock =_create_model(unit_model_def, adapter_constructor, flowsheet_manager)

    # Add the model to the flowsheet
    component_name = f"{unit_model_def.name}_{unit_model_def.id}"
    flowsheet_manager.model.fs.add_component(component_name, idaes_model)

    # Import initial guesses
    initial_values = unit_model_def.initial_values

    if initial_values:
        from_json(idaes_model, initial_values, wts=StoreSpec.value())
    
    # Fix properties
    block_context = BlockContext(flowsheet_manager.model.fs)
    _fix_properties(idaes_model,unit_model_def, flowsheet_manager,block_context)
    # Fix ports
    _fix_ports(idaes_model, unit_model_def, flowsheet_manager,block_context)
    # Apply degrees of freedom (add constraints for controlled vars)
    block_context.apply_elimination()
    return idaes_model

def _create_model(unit_model_def : UnitModelSchema,adapter_constructor: UnitModelConstructor,flowsheet_manager : FlowsheetManager) -> UnitModelBlock:
    """
    Create the kwargs for the model constructor from the unit model definition.
    """
    arg_parsers = adapter_constructor.arg_parsers
    args = unit_model_def.args

    kwargs : dict[str, Any] = {}
    for name in args:
        if not name in arg_parsers:
            raise ValueError(
                f"Argument {name} not found in model schema, available arguments are {[x for x in arg_parsers.keys()]}"
            )
    for name in arg_parsers:
        register = arg_parsers[name]
        # try to get the argument from the schema passed in
        # however, in some cases the argument may not be required
        # i.e when specified by the model adapter itself
        # e.g methods.constant

        # note that in the future we may have to support passing
        # optional arguments to the model constructor
        result = register.run(args.get(name, None),flowsheet_manager)
        kwargs[name] = result

    idaes_model = adapter_constructor.model_constructor(**kwargs)
    return idaes_model

def _fix_properties( unit_model: UnitModelBlock, unit_model_def: UnitModelSchema, flowsheet_manager : FlowsheetManager,block_context: BlockContext) -> None:
        """
        Fix the properties of the unit model based on the properties in the unit model definition.
        """
        # Loop through the properties in the unit model definition to fix the properties in the unit model
        properties = unit_model_def.properties
        fix_block(
            unit_model, properties, flowsheet_manager.model.fs, block_context
        )


def _fix_ports(unit_model: UnitModelBlock, unit_model_def: UnitModelSchema, flowsheet_manager:FlowsheetManager,block_context: BlockContext) -> None:
        """
        Fix the ports of the unit model based on the ports in the unit model definition.
        """
        # Loop through the ports in the unit model definition to fix the ports in the unit model
        for port_name, port_schema in unit_model_def.ports.items():
            # Get the port from the unit model
            port = get_component(unit_model, port_name)
            if not isinstance(port, Port):
                raise ValueError(f"Port {port_name} not found in model")
            # Register the port, so arcs can connect to it by id
            flowsheet_manager.ports.register_port(port_schema.id, port)
            # Set the port parameters
            sb = _get_state_from_port(port, 0)
            state_block = sb.parent_component()

            # Prefer time-only properties_in/properties_out if available
            # decide by port name containing "inlet"/"outlet" and "hot"/"cold"
            pn = port_name.lower()
            if "inlet" in pn:
                if "hot" in pn and hasattr(unit_model, "hot_side") and hasattr(unit_model.hot_side, "properties_in"):
                    state_block = unit_model.hot_side.properties_in
                elif "cold" in pn and hasattr(unit_model, "cold_side") and hasattr(unit_model.cold_side, "properties_in"):
                    state_block = unit_model.cold_side.properties_in
            elif "outlet" in pn:
                if "hot" in pn and hasattr(unit_model, "hot_side") and hasattr(unit_model.hot_side, "properties_out"):
                    state_block = unit_model.hot_side.properties_out
                elif "cold" in pn and hasattr(unit_model, "cold_side") and hasattr(unit_model.cold_side, "properties_out"):
                    state_block = unit_model.cold_side.properties_out

            if sb.config.defined_state:
                """
                ie. Inlet state.

                The inlet state needs a separate context, because its variables
                are all fixed during initialization, so applying elimination
                involving variables outside the inlet state would run into
                degrees of freedom issues.
                """
                inlet_ctx = BlockContext(flowsheet_manager.model.fs)
                fix_block(
                    state_block,
                    port_schema.properties,
                    flowsheet_manager.model.fs,
                    inlet_ctx,
                )
                inlet_ctx.apply_elimination()
            else:
                # The outlet state(s) can use the unit model context.
                fix_block(
                    state_block,
                    port_schema.properties,
                    flowsheet_manager.model.fs,
                    block_context,
                )

    


from typing import Any, cast, NewType
from pyomo.environ import value
from pyomo.core.base.units_container import units, _PyomoUnit
from pyomo.core.base.var import Var
from pyomo.core.base.expression import Expression, ExpressionData
from pyomo.core.expr import ExpressionBase, NPV_ProductExpression
from pint import UnitRegistry

ValueWithUnits = NewType("ValueWithUnits", NPV_ProductExpression)


def _get_pint_unit(unit: str) -> Any:
    """
    Get the pint unit object
    """
    pint_unit = getattr(units.pint_registry, unit, None)
    if pint_unit is None:
        raise AttributeError(f"Unit `{unit}` not found.")
    return pint_unit


def get_unit(unit: str) -> _PyomoUnit:
    """
    Get the pint unit object
    @unit: str unit type
    @return: unit object
    """
    return _PyomoUnit(_get_pint_unit(unit), units.pint_registry)


def attach_unit(value: float, unit: str | None) -> ValueWithUnits:
    """
    Attach the unit to the value
    @value: float value
    @unit: str unit
    @return: value with unit attached
    """
    value, unit = idaes_specific_convert(
        value, unit
    )  # make sure the value is in a unit that idaes supports
    if unit is None:
        return value
    pyomo_unit = get_unit(unit)
    return value * pyomo_unit


def check_units_equivalent(unit1: _PyomoUnit, unit2: _PyomoUnit) -> bool:
    """
    Check if two units are equivalent
    """
    if unit1 is None:
        unit1 = units.dimensionless
    if unit2 is None:
        unit2 = units.dimensionless
    return (
        unit1._get_pint_unit().dimensionality
        == unit2._get_pint_unit().dimensionality
    )



def idaes_specific_convert(value: float, unit: str | None) -> tuple[float, str | None]:
    """
    Convert the value to a unit that is specific to idaes
    (ie. idaes only supports K for temperature)
    @value: float value
    @unit: str unit
    @return: tuple:
        - (float) converted value
        - (str) new unit
    """
    if unit in ["degC", "degR", "degF"]:  # temperature units
        from_quantity = units.pint_registry.Quantity(value, unit)
        to_unit = units.pint_registry.K
        # can probably do the conversion/attachment in one step but haven't figured out how yet
        converted_value = from_quantity.to(to_unit)  # pint.Quantity object
        return converted_value.magnitude, "K"
    if unit in ["percent"]:  # dimensionless units
        from_quantity = units.pint_registry.Quantity(value, unit)
        to_unit = units.pint_registry.dimensionless
        converted_value = from_quantity.to(to_unit)
        return converted_value.magnitude, None
    if unit in [None, ""]:
        return value, None
    return value, unit


def get_attached_unit(
    var: Var | Expression | ExpressionBase | ExpressionData | float,
) -> _PyomoUnit | None:
    """
    Get the unit of a variable.
    """
    if isinstance(var, float):
        # no attached unit
        return units.dimensionless
    if isinstance(var, ExpressionData):
        var = var.parent_component()
    if var.is_indexed():
        # we will get the unit from the first item in the indexed variable
        var = var[next(iter(var.index_set()))]
    if isinstance(var, (ExpressionBase, Expression, ExpressionData)):
        # handle expressions
        return units.get_units(var)
    else:
        # handle variables
        return var.get_units()


def get_attached_unit_str(var: Var | Expression | ExpressionBase) -> str:
    """
    Get the unit of a variable as a string.
    """
    unit = get_attached_unit(var)
    return str(unit) if unit is not None else "dimensionless"


def get_value(var: Var | Expression | ExpressionBase) -> float | dict:
    if var.is_indexed():
        if isinstance(var, Var):
            # can get values directly
            return cast(dict, var.get_values())
        else:
            data = {}
            for index in var.index_set():
                data[index] = value(var[index])
            return data
    else:
        return float(value(var))
